# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
HL7apy - core classes
"""

import collections
import datetime
from itertools import takewhile
import importlib

from hl7apy import get_default_version, get_default_encoding_chars, \
                   get_default_validation_level, check_validation_level, \
                   check_encoding_chars, check_version, load_library, \
                   find_reference, load_reference
from hl7apy.validation import Validator
from hl7apy.exceptions import ChildNotFound, ChildNotValid, \
                              MaxChildLimitReached, OperationNotAllowed, \
                              InvalidName
from hl7apy.factories import datatype_factory
from hl7apy.base_datatypes import BaseDataType

def is_base_datatype(datatype, version=None):
    """
    Check if the given datatype is a base datatype of the specified version

    :type datatype: ``basestring``
    :param datatype: the datatype (e.g. ST)
    :param version: the HL7 version (e.g. 2.5)
    :return: ``True`` if it is a base datatype, ``False`` otherwise

    >>> is_base_datatype('ST')
    True
    >>> is_base_datatype('CE')
    False
    """
    if version is None:
        version = get_default_version()
    lib = load_library(version)
    return lib.is_base_datatype(datatype)


def _remove_trailing(children):
    trailing = list(takewhile(lambda x: not x, reversed(children)))
    if len(trailing) > 0:
        children = children[:-len(trailing)]
    return children


def _valid_child_name(child_name, expected_parent):
    try:
        parent, index = child_name.split("_")
        int(index)
    except (ValueError, AttributeError):
        return False
    else:
        if str(parent).upper() != str(expected_parent).upper():
            return False
        return True


class ElementProxy(collections.Sequence):
    """
    It contains the results of a child traversal, and provides lazy child instantiation
    in order to support the following API:

    >>> m = Message("OML_O33")
    >>> print m.msh
    [<Segment MSH>]
    >>> print m.msh.__class__
    <class '__main__.ElementProxy'>
    >>> print m.msh.msh_7.__class__
    <class '__main__.ElementProxy'>
    """
    cls_attrs = ('element_list', 'list', 'element_name')

    def __init__(self, element_list, element_name, elements):
        self.element_name = element_name
        self.element_list = element_list
        self.list = list(elements)

    def __len__(self):
        return len(self.list)

    def __iter__(self):
        if not self.list:
            return iter([])
        else:
            return super(ElementProxy, self).__iter__()

    def __getattr__(self, name):
        try:
            element = self.list[0]
        except IndexError: # first child not found, create the element
            element = self.element_list.create_element(self.element_name, traversal_parent=True)
        return getattr(element, name)

    def __setattr__(self, name, value):
        if name in self.cls_attrs:
            super(ElementProxy, self).__setattr__(name, value)
        else:
            if name == 'value':
                element = self.element_list.element
                name = self.element_name
            else:
                try:
                    element = self.list[0]
                except IndexError: # first child not found, create the element
                    element = self.element_list.create_element(self.element_name, traversal_parent=True)
            setattr(element, name, value)

    def __setitem__(self, index, value):
        self.element_list.set(self.element_name, value, index)

    def __getitem__(self, index):
        return self.list[index]

    def __delitem__(self, index):
        self.element_list.remove(self.list[index])
        del self.list[index]

    def __delattr__(self, name):
        delattr(self.list[0], name)

    def __repr__(self):
        return str(self.list)


class ElementList(collections.MutableSequence):
    """
    Delegate for handling the children of a given Element.

    It stores the children inside a list and indexes them using a dictionary (e.g. self.indexes['SPM'] will return
    all child named 'SPM' found in the Message instance)

    """
    def __init__(self, element):
        self.element = element
        self.list = []
        self.indexes = {}

    def get_ordered_children(self):
        """
        Return the list of children ordered according to the element structure

        :return: a list of :class:`hl7apy.core.Element`
        """
        ordered_keys = self.element.structure_by_name.keys() if self.element.structure_by_name is not None else []
        children = [self.indexes.get(k, None) for k in ordered_keys]
        return _remove_trailing(children)

    def get_children(self):
        """
        Return the list of children according to their insertion order

        :return: a list of :class:`hl7apy.core.Element`
        """
        return [(v,) for v in self.list]

    def insert(self, index, child, by_name_index=-1):
        """
        Add the child at the given index

        :type index: ``int``
        :param index: child position
        :param child: an instance of an :class:`hl7apy.core.Element` subclass
        """
        if self._can_add_child(child):
            try:
                if by_name_index == -1:
                    self.indexes[child.name].append(child)
                else:
                    self.indexes[child.name].insert(by_name_index, child)
            except KeyError:
                self.indexes[child.name] = [child]
            self.list.insert(index, child)

    def append(self, child):
        """
        Append the given child

        :param child: an instance of an :class:`hl7apy.core.Element` subclass
        """
        if self._can_add_child(child):
            self.list.append(child)
            try:
                self.indexes[child.name].append(child)
            except KeyError:
                self.indexes[child.name] = [child]

    def pop(self, index=0):
        child = super(ElementList, self).pop(index)
        self._remove_from_index(child)
        return child

    def get(self, name):
        """
        Get the children having the given name

        :param name: the name of the children (e.g. PID)
        :return: an instance of :class:`hl7apy.core.ElementProxy` containing the results
        """
        return self._default_child_lookup(name)

    def set(self, name, value, index=-1):
        """
        Assign the ``value`` to the child having the given ``name`` at the ``index`` position

        :type name: ``basestring``
        :param name: the child name (e.g. PID)
        :type value: an instance of :class:`hl7apy.core.Element`, a `basestring` or an instance of :class:`hl7apy.core.ElementProxy`
        :param value: the child value
        :param index: the child position (e.g. 1)
        """
        if isinstance(value, ElementProxy): # just copy the first element of the ElementProxy (e.g. message.pid = message2.pid)
            value = value[0].to_er7()

        name = name.upper()
        reference = None if name is None else self.element.find_child_reference(name)
        child_ref, child_name = (None, None) if reference is None else (reference['ref'], reference['name'])

        if isinstance(value, basestring): # if the value is a basestring, parse it
            child = self.element.parse_child(value, child_name=child_name, reference=child_ref)
        elif isinstance(value, Element): # it is already an instance of Element
            child = value
        else:
            raise ChildNotValid(value, child_name)

        if child.name != child_name: # e.g. message.pid = Segment('SPM') is forbidden
            raise ChildNotValid(value, child_name)

        child_to_remove = self.child_at_index(child_name, index)

        if child_to_remove is None:
            self.append(child)
        else:
            self.replace_child(child_to_remove, child)

        # a set has been called, change the temporary parent to be the actual one
        self.element.set_parent_to_traversal()

    def remove(self, child):
        """
        Remove the given child from both child list and child indexes

        :param child: an instance of :class:`hl7apy.core.Element` subclass
        """
        self._remove_from_index(child)
        self.list.remove(child)

    def remove_by_name(self, name, index=0):
        """
        Remove the child having the given name at the given position

        :param name: child name (e.g. PID)
        :param index: child index
        :return: an instance of :class:`hl7apy.core.Element` subclass
        """
        child = self.child_at_index(name, index)
        self.remove(child)
        return child

    def child_at_index(self, name, index):
        """
        Return the child named `name` at the given index

        :param name: child name (e.g. PID)
        :param index: child index
        :return: an instance of :class:`hl7apy.core.Element` subclass
        """
        child_name = None if name is None else self._find_name(name)
        try:
            child = self.indexes[child_name][index]
        except (KeyError, IndexError):
            child = None
        return child

    def replace_child(self, old_child, new_child):
        list_index = self.list.index(old_child)
        by_name_index = self.indexes[old_child.name].index(old_child)
        self.remove(old_child)
        self.insert(list_index, new_child, by_name_index)

    def create_element(self, name, traversal_parent=False, reference=None):
        """
        Create an element having the given name

        :param name: the name of the element to be created (e.g. PID)
        :param traversal_parent: if True, the parent will be set as temporary for traversal purposes
        :param reference: the new element structure (see :func:`hl7apy.load_reference`)
        :return: an instance of an :class:`hl7apy.core.Element` subclass
        :raises: :exc:`hl7apy.exceptions.ChildNotFound` if the element does not exist
        """
        if reference is None:
            name = name.upper()
            reference = self.element.find_child_reference(name)
        if reference is not None:
            cls = reference['cls']
            element_name = reference['name']
            kwargs = {'reference': reference['ref'], 'validation_level': self.element.validation_level}
            if not traversal_parent:
                kwargs['parent'] = self.element
            else:
                kwargs['traversal_parent'] = self.element
            return cls(element_name, **kwargs)
        else:
            raise ChildNotFound(name)

    def _find_name(self, name):
        """
        Find the reference of a child having the given name

        :param name: the child name (e.g. PID)
        :return: the element structure (see :func:`hl7apy.load_reference`) or `None` if the element has not been found
        """
        name = name.upper()
        element = self.element.find_child_reference(name)
        return element['name'] if element is not None else None

    def _default_child_lookup(self, name):
        """
        Return an instance of :class:`hl7apy.core.ElementProxy` containing the children found having the given name

        :param name: the name of the children (e.g. PID)
        :return: an instance of :class:`hl7apy.core.ElementProxy` containing the results
        """
        if self.indexes.has_key(name): # use the indexes to find the children faster
            results = ElementProxy(self, name, (c for c in self.indexes[name]))
            return results
        else: # the child has not been found in the indexes dictionary (e.g. msh_9.message_code, msh_9.msh_9_1)
            child_name = self._find_name(name)
            if child_name is not None:
                results = ElementProxy(self, name, (c for c in self.indexes.get(child_name, [])))
                return results

    def _can_add_child(self, child):
        if self.element._is_valid_child(child):
            if child.parent != self.element: #avoid infinite recursion
                child.parent = self.element
            else:
                # if validation is strict, check the child cardinality
                if Validator.is_strict(self.element.validation_level):
                    min, max = self.element.repetitions.get(child.name, (0, -1))
                    if len(self.indexes.get(child.name, [])) + 1 > int(max) and max > -1:
                        raise MaxChildLimitReached(self.element, child, max)
                return True
        else:
            raise ChildNotValid(child, self.element)
        return False

    def _remove_from_index(self, child):
        try:
            self.indexes[child.name].remove(child)
        except (KeyError, ValueError):
            pass

    def __len__(self):
        return len(self.list)

    def __getitem__(self, index):
        return self.list[index]

    def __delitem__(self, index):
        child = self.list[index]
        self._remove_from_index(child)
        del self.list[index]

    def __setitem__(self, index, value):
        child_name = self.list[index].name
        self.set(child_name, value, index)

    def __str__(self):
        return str(self.list)


class ElementFinder(object):

    @staticmethod
    def get_structure(element, reference=None):
        """
        Get the element structure

        :param element: element having the given reference structure
        :param reference: the element structure (see :func:`hl7apy.load_reference)
        :return: a dictionary containing the structure data
        """
        if reference is None:
            try:
                reference = load_reference(element.name, element.classname, element.version)
            except (ChildNotFound, KeyError):
                raise InvalidName(element.classname, element.name)
        if not isinstance(reference, collections.Sequence):
            raise Exception
        return ElementFinder._parse_structure(element, reference)

    @staticmethod
    def _parse_structure(element, reference):
        """
        Parse the given reference

        :param element: element having the given reference structure
        :param reference: the element structure (see :func:`hl7apy.load_reference)
        :return: a dictionary containing the structure data
        """
        content_type = reference[0] # content type can be sequence, choice or leaf
        data = {}
        if content_type in ('sequence', 'choice'):
            children = reference[1]
            structure = collections.OrderedDict()
            repetitions = {child_name:cardinality for child_name, cardinality in children}
            for c in children:
                child_name, cardinality = c
                structure[child_name] = find_reference(child_name, element.child_classes, element.version)
            data['structure_by_longname'] = {e['ref'][2]: e for e in structure.values() if e['ref'][0] == 'leaf'}
            data['structure_by_name'] = structure
            data['repetitions'] = repetitions
        elif content_type == 'leaf':
            child_type, datatype, long_name, table = reference
            data['datatype'] = datatype
            data['table'] = table
            data['long_name'] = long_name
            if not is_base_datatype(data['datatype'], element.version) and \
                    data['datatype'] != 'varies':
                try:
                    reference = find_reference(data['datatype'], element.child_classes, element.version)
                    data.update(ElementFinder._parse_structure(element, reference['ref']))
                except ChildNotFound: # if element.child_classes is empty (e.g SubComponent)
                    pass

        return data


class Element(object):
    """
    Base class for all HL7 elements. It is not meant to be directly instantiated.
    """

    cls_attrs = ['name', 'validation_level', 'version', 'children',
                 'table', 'long_name', 'value', '_value', 'parent',  '_parent', 'traversal_parent',
                 'child_classes', 'encoding_chars', 'structure_by_name', 'structure_by_longname', 'repetitions']

    def __init__(self, name=None, parent=None, reference=None, version=None,
                 validation_level=None, traversal_parent=None):

        if self.__class__ == Element:
            raise OperationNotAllowed("Cannot instantiate an Element")

        if validation_level is None:
            validation_level = get_default_validation_level()

        if version is None:
            version = get_default_version()

        check_validation_level(validation_level)
        check_version(version)

        self.validation_level = validation_level
        self.name = name.upper() if name is not None else None
        self.version = version
        self.table = None
        self.long_name = None
        self.children = ElementList(self)
        self.parent = parent
        if parent is None:
            self.traversal_parent = traversal_parent
        else:
            self.traversal_parent = None
        self.structure_by_name = None
        self.structure_by_longname = None
        self.repetitions = {}
        if self.name is not None:
            structure = ElementFinder.get_structure(self, reference)
            for k, v in structure.iteritems():
                setattr(self, k, v)

    def find_child_reference(self, name):
        name = name.upper()
        if isinstance(self.structure_by_name, collections.MutableMapping):
            element =  self.structure_by_name.get(name) or self.structure_by_longname.get(name)
        else:
            element = None
        if element is None: # not found in self.structure
            element = find_reference(name, self.child_classes, self.version)
            if Validator.is_strict(self.validation_level): # cannot be created if validation is strict
                raise ChildNotValid(name, self)
        return element

    def add(self, obj):
        """
        Add an instance of :class:`hl7apy.core.Element` subclass to the list of children

        :param obj: an instance of :class:`hl7apy.core.Element` subclass

        >>> s = Segment('PID')
        >>> f = Field('PID_5')
        >>> f.value = 'EVERYMAN^ADAM'
        >>> s.add(f)
        >>> print s.to_er7()
        PID|||||EVERYMAN^ADAM
        """
        self.children.append(obj)

    def is_named(self, name):
        name = name.upper()
        return name in (self.name, self.long_name)

    def is_unknown(self):
        return self.name is None

    def set_parent_to_traversal(self):
        if self.traversal_parent and self.parent is None:
            self.parent = self.traversal_parent
            self.parent.set_parent_to_traversal()
        else:
            self.traversal_parent = None

    def parse_child(self, text, **kwargs):
        if self.child_parser:
            kwargs['version'] = self.version
            kwargs['validation_level'] = self.validation_level
            module = importlib.import_module("hl7apy.parser")
            parser = getattr(module, self.child_parser[0])
            return parser(text, **kwargs)
        return None

    def parse_children(self, text, **kwargs):
        if self.child_parser:
            kwargs['version'] = self.version
            kwargs['validation_level'] = self.validation_level
            module = importlib.import_module("hl7apy.parser")
            parser = getattr(module, self.child_parser[1])
            return parser(text, **kwargs)

    def to_er7(self, encoding_chars=None, trailing_children=False):
        """
        Returns the HL7 representation of the :class:`Element`. It adds the appropriate
        separator at the end if needed

        :param encoding_chars: The encoding chars to use.
            If it is ``None`` it uses :attr:`self.encoding_chars`,
            which by default is :attr:`hl7apy.consts.DEFAULT_ENCODING_CHARS`

        :rtype: ``str`` the HL7 representation of the :class:`Element`
        """
        if encoding_chars is None:
            encoding_chars = self.encoding_chars

        child_class = self.child_classes[0]
        separator = encoding_chars.get(child_class.__name__.upper(), '')

        s = []
        for child in self._get_children(trailing_children):
            if child:
                s.extend(repetition.to_er7(encoding_chars) for repetition in child)
            else:
                try:
                    s.append(self._handle_empty_children(encoding_chars))
                except NotImplementedError:
                    pass


        return separator.join(s)

    def validate(self):
        """
        Validate the HL7 element using the :attr:`hl7apy.consts.VALIDATION_LEVEL.STRICT` validation level

        :rtype: ``bool``
        :return: True if validation succeeds, False otherwise
        """
        return Validator.validate(self)

    def _get_parent(self):
        return self._parent

    def _set_parent(self, parent):
        self._parent = parent
        self.traversal_parent = None
        if parent is not None:
            self.parent.add(self)

    parent = property(_get_parent, _set_parent,
                      doc="The parent :class:`Element` of this one")

    def _set_value(self, value):
        self.children = self.parse_children(value)

    def _get_value(self):
        return self.to_er7()

    value = property(_get_value, _set_value)

    @property
    def classname(self):
        """
        The name of the class
        """
        return self.__class__.__name__

    @property
    def encoding_chars(self):
        """
        A ``dict`` with the encoding chars of the :class:`Element`.
        If the :class:`Element` has a parent it is the parent's
        ``encoding_chars`` otherwise the :attr:`consts.DEFAULT_ENCODING_CHARS`
        The structure of the ``dict`` is:

        .. code-block:: python

            {'SEGMENT' : '\\r',
            'GROUP': '\\r',
            'FIELD' : '|',
            'COMPONENT' : '^',
            'SUBCOMPONENT' : '&',
            'REPETITION' : '~',
            'ESCAPE' : '\\'}

        """
        if self.parent is not None:
            return self.parent.encoding_chars
        return get_default_encoding_chars()

    def _is_valid_child(self, child):
        valid = child.classname in (c.__name__ for c in self.child_classes)
        if valid:
            if child.name is not None:
                self.find_child_reference(child.name)
        return valid

    def _get_children(self, trailing=False):
        children = self.children.get_ordered_children()
        children.extend([c for c in self.children.get_children() if c[0].name in (None, 'ST')])
        if not trailing:
            children = _remove_trailing(children)
        return children

    def _handle_empty_children(self, encoding_chars=None):
        """
        Subclasses should implement this method, which must returns the string
        that should be used when a child of the structure is empty
        """
        raise NotImplementedError

    def __getattr__(self, name):
        if hasattr(self, 'children'):
            return self.children.get(name)
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if name in self.cls_attrs:
            children = []
            if name == 'children':
                if not isinstance(value, ElementList):
                    children = value
                    value = ElementList(self)
            super(Element, self).__setattr__(name, value)
            for c in children:
                self.add(c)
        elif hasattr(self, 'children'):
            self.children.set(name, value, 0)

    def __delattr__(self, name):
        if name in self.cls_attrs:
            super(Element, self).__delattr__(name)
        else:
            self.children.remove_by_name(name)

    def __repr__(self):
        return "<{0} {1}>".format(self.classname, self.name or '')


class SupportComplexDataType(Element):
    """
    Mixin for classes that support complex datatypes
    """

    def __init__(self):

        if self.__class__ == SupportComplexDataType:
            raise OperationNotAllowed("Cannot instantiate a SupportComplexDataType")

        self.cls_attrs.extend(['_datatype', 'datatype'])
        self._datatype = None

    def is_unknown(self):
        return self.name == self.datatype

    def _get_datatype(self):
        return self._datatype

    def _set_datatype(self, datatype):
        if Validator.is_strict(self.validation_level) and self.datatype and \
                datatype != self.datatype:
            raise OperationNotAllowed("Cannot change datatype using STRICT validation")

        if self.name is None and not is_base_datatype(datatype, self.version) and datatype is not None:
            reference = load_reference(datatype, self.classname, self.version)
            structure = ElementFinder.get_structure(self, reference)
            for k, v in structure.iteritems():
                setattr(self, k, v)

        if hasattr(self, 'children') and len(self.children) >= 1:
            if is_base_datatype(self.datatype, self.version):
                self._datatype = datatype
                if is_base_datatype(datatype, self.version):
                    self.children[0].datatype = datatype
            else:
                raise OperationNotAllowed("Cannot change datatype, since the Element already contains children")
        else:
            self._datatype = datatype

    datatype = property(_get_datatype, _set_datatype)

    def _set_value(self, value):
        children = self.parse_children(value)
        if Validator.is_quiet(self.validation_level) and is_base_datatype(self.datatype, self.version) and \
                len(children) > 1:
            self.datatype = None
        self.children = children

    def _get_value(self):
        return self.to_er7()

    value = property(_get_value, _set_value)

    def _get_children(self, trailing=False):
        if is_base_datatype(self.datatype, self.version) or self.datatype is None:
            return [[c for c in self.children]]
        else:
            return Element._get_children(self, trailing=False)

    def _is_valid_child(self, child):
        if not is_base_datatype(self.datatype, self.version):
            if self.datatype == 'varies' and _valid_child_name(child.name, self.datatype):
                return True
            if self.datatype is None and _valid_child_name(self.name, 'varies') and child.is_unknown():
                return True
            if child.is_unknown() and Validator.is_strict(self.validation_level): # cannot add unknown children in strict validation
                return False
            if not child.is_unknown() and self.datatype and not _valid_child_name(child.name, self.datatype):
                return False
        try:
            valid = Element._is_valid_child(self, child)
            if valid and is_base_datatype(self.datatype, self.version):
                if child.datatype and child.datatype != self.datatype:
                    valid = False
            return valid
        except ChildNotFound:
            if is_base_datatype(self.datatype, self.version):
                if child.datatype and child.datatype != self.datatype:
                    return False
            return True

    def _handle_empty_children(self, encoding_chars=None):
        return ''

    def __repr__(self):
        if self.name:
            return '<{0} {1} ({2}) of type {3}>'.format(self.classname, self.name, self.long_name,
                                            self.datatype)
        else:
            return '<{0} of type {1}>'.format(self.classname, self.datatype)


class CanBeVaries(Element):
    """
    Mixin for Elements that can be of VARIES datatype
    """
    def __init__(self, name=None, datatype=None, parent=None, reference=None,
                 version=None, validation_level=None, traversal_parent=None):

        if self.__class__ == CanBeVaries:
            raise OperationNotAllowed("Cannot instantiate a CanBeVaries")

        if datatype == 'varies' and reference is None:
            reference = ('leaf', 'varies', None, None)

        if name is not None and _valid_child_name(name, 'VARIES'):
            Element.__init__(self, None, None, reference, version,
                                            validation_level, traversal_parent)
            self.name = name.upper()
        else:
            Element.__init__(self, name, None, reference, version,
                                            validation_level, traversal_parent)

        # it means that the name was found in the references but it's not a valid reference (e.g. SubComponent('CX'))
        if self.name and not self.name.startswith('VARIES') and self.datatype is None:
            raise InvalidName(self.classname, self.name)

        if self.name: # it means that the datatype has been valued by the finder
            # if it is STRICT we don't allow to change the official datatype
            #TODO: should check if it is VARIES
            if Validator.is_strict(self.validation_level) and not None in (datatype, self.datatype) and \
                    datatype != self.datatype:
                raise OperationNotAllowed("Cannot override datatype in strict mode")
            # in QUIET we overwrite it only if the given one is not None
            elif datatype is not None:
                self.datatype = datatype
        else:
            self.datatype = datatype
            self.name = self.datatype

        self.parent = parent
        if parent is None:
            self.traversal_parent = traversal_parent
        else:
            self.traversal_parent = None

    def is_unknown(self):
        return self.name == self.datatype


class SubComponent(CanBeVaries):
    """
    Class representing an HL7 subcomponent.

    :type name: ``basestring``
    :param name: the HL7 name of the subcomponent (e.g. CWE_1)

    :type datatype: ``basestring``
    :param datatype: the datatype of the component (e.g. ST)

    :type value: ``basestring`` or instance of :class:`hl7apy.base_datatypes.BaseDataType`
    :param value: the value of the subcomponent (e.g. ADT_A01)

    :type parent: an instance of :class:`hl7apy.core.Component` or None
    :param parent: the parent

    :param reference: the reference structure (see :func:`hl7apy.load_reference`)

    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)

    :type validation_level: ``int``
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)

    :type traversal_parent: an instance of :class:`hl7apy.core.Message`, :class:`hl7apy.core.Group` or None
    :param traversal_parent: the temporary parent used during traversal
    """
    child_classes = ()

    def __init__(self, name=None, datatype=None, value=None, parent=None,
                 reference=None, version=None, validation_level=None,
                 traversal_parent=None):
        self.cls_attrs.extend([ '_value', 'value', 'datatype', '_datatype'])
        self._parent = None # we need to initialize it
        self._datatype = None # we need to initialize it
        if not name and datatype is None:
            raise OperationNotAllowed("Cannot instantiate a SubComponent with name and datatype both empty")

        CanBeVaries.__init__(self, name, datatype, parent, reference,
                 version, validation_level, traversal_parent)

        if _valid_child_name(name, 'VARIES') and self.datatype is None:
            self.datatype = 'ST'

        self.value = value

    def add(self, obj):
        raise OperationNotAllowed("Cannot add children to a SubComponent")

    def to_er7(self, encoding_chars=None, trailing_chars=None):
        """
        Return the ER7-encoded string

        :type encoding_chars: ``dict``
        :param encoding_chars: a dictionary containing the encoding chars or None to use the default (see :func:`hl7apy.set_default_encoding_chars`)

        :type trailing_children: ``bool``
        :param trailing_children: if True, trailing children will be added even if their value is None

        :return: the ER7-encoded string

        >>> s = SubComponent("CE_1")
        >>> s.value = "IDENTIFIER"
        >>> print s.to_er7()
        IDENTIFIER
        """

        if encoding_chars is None:
            encoding_chars = self.encoding_chars
        try:
            return self.value.to_er7(encoding_chars)
        except AttributeError:
            return self.value

    def _set_value(self, value):
        if value is None:
            self._value = None
        else:
            if value and isinstance(value, basestring):
                self._value = datatype_factory(self.datatype, value, self.version,
                                               self.validation_level)
            elif not value or isinstance(value, BaseDataType):
                self._value = value
            else:
                raise ValueError('Cannot assign {0}'.format(value.classname))
            self.set_parent_to_traversal()

    def _get_value(self):
        return self._value

    value = property(_get_value, _set_value)

    def _get_datatype(self):
        return self._datatype

    def _set_datatype(self, datatype):

        if datatype and not is_base_datatype(datatype, self.version): # only base datatypes allowed
            raise OperationNotAllowed("Cannot set a complex datatype to a SubComponent")

        if Validator.is_strict(self.validation_level) and \
                self.datatype is not None and datatype != self.datatype:
            raise OperationNotAllowed("Cannot change datatype using STRICT validation")
        # if the SC is already valued it is impossible to change the datatype
        if hasattr(self, 'value') and self.value:
            raise OperationNotAllowed("Cannot change datatype: the Subcomponent is valued")
        else:
            #change also the parent datatype if it is a base datatype
            if self.parent is not None and \
                is_base_datatype(self.parent.datatype, self.parent.version) and \
                self.parent.datatype != datatype : # avoid infinite recursion
                self.parent.datatype = datatype
            self._datatype = datatype

    datatype = property(_get_datatype, _set_datatype)

    def __repr__(self):
        return '<SubComponent {0}>'.format(self.name or self.datatype)


class Component(SupportComplexDataType, CanBeVaries):
    """
    Class representing an HL7 component.

    :type name: ``basestring``
    :param name: the HL7 name of the component (e.g. XPN_2)

    :type datatype: ``basestring``
    :param datatype: the datatype of the component (e.g. CE)

    :type parent: an instance of :class:`hl7apy.core.Field` or None
    :param parent: the parent

    :param reference: the reference structure (see :func:`hl7apy.load_reference`)

    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)

    :type validation_level: ``int``
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)

    :type traversal_parent: an instance of :class:`hl7apy.core.Message`, :class:`hl7apy.core.Group` or None
    :param traversal_parent: the temporary parent used during traversal
    """
    child_classes = (SubComponent,)
    child_parser = ('parse_subcomponent', 'parse_subcomponents')

    def __init__(self, name=None, datatype=None, parent=None, reference=None,
                 version=None, validation_level=None, traversal_parent=None):

        SupportComplexDataType.__init__(self)

        if datatype == 'varies' and reference is None:
            reference = ('leaf', 'varies', None, None)

        CanBeVaries.__init__(self, name, datatype, parent, reference,
                 version, validation_level, traversal_parent)

        if self.is_unknown() and Validator.is_strict(validation_level) and \
                not is_base_datatype(self.datatype, self.version) and self.datatype != 'varies':
            raise OperationNotAllowed("Cannot instantiate an unknown Element with strict validation")

        #TODO: This control should be deleted (see CanBeVaries)
        if datatype is not None and Validator.is_strict(validation_level) and self.datatype != 'varies' and self.datatype != datatype:
            raise OperationNotAllowed("Cannot assign a different datatype with strict validation")

    def add_subcomponent(self, name):
        """
        Create an instance of :class:`hl7apy.core.SubComponent` having the given name

        :param name: the name of the subcomponent to be created (e.g. CE_1)
        :return: an instance of :class:`hl7apy.core.SubComponent`

        >>> c = Component(datatype='CE')
        >>> ce_1 = c.add_subcomponent('CE_1')
        >>> print ce_1
        <SubComponent CE_1>
        >>> print ce_1 in c.children
        True
        """
        if self.is_unknown() and is_base_datatype(self.datatype):
            # An unknown component can't have a child
            raise ChildNotValid(name, self)
        return self.children.create_element(name)

    def add(self, obj):
        """
        Add an instance of :class:`hl7apy.core.SubComponent` to the list of children

        :param obj: an instance of :class:`hl7apy.core.SubComponent`

        >>> c = Component('CX_10')
        >>> s = SubComponent(name='CWE_1', value='EXAMPLE_ID')
        >>> s2 = SubComponent(name='CWE_4', value='ALT_ID')
        >>> c.add(s)
        >>> c.add(s2)
        >>> print c.to_er7()
        EXAMPLE_ID&&&ALT_ID
        """
        # base datatype components can't have more than one child
        if self.name and is_base_datatype(self.datatype, self.version) and \
                len(self.children) >= 1:
            raise MaxChildLimitReached(self, obj, 1)

        # if the name is different from the datatype (i.e. the name has been forced to be the same as the datatype)
        if obj.name and obj.name != obj.datatype:
            try:
                if not _valid_child_name(obj.name, self.datatype):
                    raise ChildNotValid(obj.name, self)
            except AttributeError:
                pass

        return super(Component, self).add(obj)

    def parse_child(self, text, child_name=None, reference=None):
        kwargs = {'name': child_name}
        if reference is not None:
            kwargs['datatype'] = reference[1]
        return super(Component, self).parse_child(text, **kwargs)

    def parse_children(self, text, **kwargs):
        kwargs = {'component_datatype' : self.datatype, 'encoding_chars' : self.encoding_chars}
        return super(Component, self).parse_children(text, **kwargs)


class Field(SupportComplexDataType):
    """
    Class representing an HL7 field.

    :type name: ``basestring``
    :param name: the HL7 name of the field (e.g. PID_5)

    :type datatype: ``basestring``
    :param datatype: the datatype of the field (e.g. CE)

    :type parent: an instance of :class:`hl7apy.core.Segment` or None
    :param parent: the parent

    :param reference: the reference structure (see :func:`hl7apy.load_reference`)

    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)

    :type validation_level: ``int``
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)

    :type traversal_parent: an instance of :class:`hl7apy.core.Message`, :class:`hl7apy.core.Group` or None
    :param traversal_parent: the temporary parent used during traversal
    """
    child_classes = (Component,)
    child_parser = ('parse_component', 'parse_components')

    def __init__(self, name=None, datatype=None, parent=None, reference=None,
                 version=None, validation_level=None, traversal_parent=None):

        SupportComplexDataType.__init__(self)

        if name is None and Validator.is_strict(validation_level) and datatype != 'varies':
            raise OperationNotAllowed("Cannot instantiate an unknown Element with strict validation")

        if datatype is not None and Validator.is_strict(validation_level) and datatype != 'varies':
            raise OperationNotAllowed("Cannot assign a different datatype with strict validation")

        if datatype == 'varies' and reference is None:
            reference = ('leaf', 'varies', None, None)

        Element.__init__(self, name, parent, reference, version,
                                    validation_level, traversal_parent)

        if datatype is not None: # force the datatype to be the one chosen by the user
            self.datatype = datatype
        elif self.name is None: # if it is unknown and no datatype has been given
            self.datatype = None

    def add_component(self, name):
        """
        Create an instance of :class:`hl7apy.core.Component` having the given name

        :param name: the name of the component to be created (e.g. XPN_2)
        :return: an instance of :class:`hl7apy.core.Component`

        >>> s = Field('PID_5')
        >>> print s.add_component('XPN_2')
        <Component XPN_2 (GIVEN_NAME) of type ST>
        """
        return self.children.create_element(name)

    def find_child_reference(self, name):
        if is_base_datatype(self.datatype, self.version):
            # create reference in case of base datatypes
            if name == self.datatype:
                element = {'cls': Component,
                           'name': self.datatype,
                           'ref': ('leaf', self.datatype, None, None)}
                return element
            raise ChildNotFound(name)
        elif self.datatype == 'varies' and _valid_child_name(name, self.datatype):
            # create reference for children in case of datatype VARIES
            element = {'cls': Component,
                       'name': name,
                       'ref' : ('leaf', None, None, None)}
            return element
        return super(Field, self).find_child_reference(name)

    def add(self, obj):
        """
        Add an instance of :class:`hl7apy.core.Component` to the list of children

        :param obj: an instance of :class:`hl7apy.core.Component`

        >>> f = Field('PID_5')
        >>> f.xpn_1 = 'EVERYMAN'
        >>> c = Component('XPN_2')
        >>> c.value = 'ADAM'
        >>> f.add(c)
        >>> print f.to_er7()
        EVERYMAN^ADAM
        """
        # base datatype components can't have more than one child
        if self.name and is_base_datatype(self.datatype, self.version) and \
                len(self.children) >= 1:
            raise MaxChildLimitReached(self, obj, 1)

        return super(Field, self).add(obj)

    def parse_child(self, text, child_name=None, reference=None):
        kwargs = {'encoding_chars': self.encoding_chars, 'reference': reference, 'name': child_name}
        if reference is not None:
            kwargs['datatype'] = reference[1]
        return super(Field, self).parse_child(text, **kwargs)

    def parse_children(self, text, **kwargs):
        kwargs = {'field_datatype': self.datatype, 'encoding_chars' : self.encoding_chars}
        return super(Field, self).parse_children(text, **kwargs)

    def to_er7(self, encoding_chars=None, trailing_children=False):
        """
        Return the ER7-encoded string

        :type encoding_chars: ``dict``
        :param encoding_chars: a dictionary containing the encoding chars or None to use the default (see :func:`hl7apy.set_default_encoding_chars`)

        :type trailing_children: ``bool``
        :param trailing_children: if True, trailing children will be added even if their value is None

        :return: the ER7-encoded string

        >>> msh_9 = Field("MSH_9")
        >>> msh_9.value = "ADT^A01^ADT_A01"
        >>> print msh_9.to_er7()
        ADT^A01^ADT_A01
        """
        if encoding_chars is None:
            encoding_chars = self.encoding_chars
        if self.is_named('MSH_1'):
            try:
                return self.msh_1_1.children[0].value.value
            except IndexError:
                return self.msh_1_1.children[0].value
        elif self.is_named('MSH_2'):
            try:
                return self.msh_2_1.children[0].value.value
            except IndexError:
                return self.msh_2_1.children[0].value
        return super(Field, self).to_er7(encoding_chars)

    def _set_value(self, value):
        if self.name in ('MSH_1', 'MSH_2'):
            s = SubComponent(datatype='ST', value=value)
            c = Component(datatype='ST')
            c.add(s)
            self.add(c)
        else:
            super(Field, self)._set_value(value)

    def _get_value(self):
        return super(Field, self)._get_value()

    value = property(_get_value, _set_value)

    def _get_children(self, trailing=False):
        if self.datatype == 'varies':
            children = [ self.children.indexes['VARIES_{0}'.format(i+1)] for i in xrange(len(self.children)) ]
            children = _remove_trailing(children)
            children.extend([[c] for c in self.children if c.is_unknown()])
            return children
        else:
            return SupportComplexDataType._get_children(self)

    def _get_traversal_children(self, name):
        """
        Retrieve component and subcomponent indexes from the given traversal path
        (e.g. PID_1_2 -> component=2, subcomponent=None)
        """
        name = name.upper()
        parts = name.split('_')
        try:
            assert 3 <= len(parts) <= 4
            prefix = "{0}_{1}".format(parts[0], parts[1])
            component = int(parts[2])
            subcomponent = int(parts[3]) if len(parts) == 4 else None
        except (AssertionError, ValueError):
            return None, None
        else:
            if prefix != self.name:
                return None, None
        return (component, subcomponent)

    def _do_traversal(self, mode, name, value=None):
        try:
            if mode == 'get':
                return super(Field, self).__getattr__(name)
            elif mode == 'set':
                super(Field, self).__setattr__(name, value)
            else:
                super(Field, self).__delattr__(name)
        except ChildNotFound:
            component, subcomponent = self._get_traversal_children(name)
            if component is None:
                raise ChildNotFound(name)

            if is_base_datatype(self.datatype, self.version):
                if subcomponent is not None or component != 1:
                    raise ChildNotFound(name)
                component_name = self.datatype
            else:
                component_name = '{0}_{1}'.format(self.datatype, component)
            if subcomponent is None:
                if mode == 'get':
                    return getattr(self, component_name)
                elif mode == 'set':
                    setattr(self, component_name, value)
                else:
                    delattr(self, component_name)
            else:
                component = getattr(self, component_name)
                component_datatype =  self.structure_by_name[component_name]['ref'][1]
                subcomponent_name = '{0}_{1}'.format(component_datatype, subcomponent)
                try:
                    if mode == 'get':
                        return getattr(component, subcomponent_name)
                    elif mode == 'set':
                        setattr(component, subcomponent_name, value)
                    else:
                        delattr(component, subcomponent_name)
                except ChildNotFound:
                    raise ChildNotFound(name)

    def __getattr__(self, name):
        return self._do_traversal('get', name)

    def __setattr__(self, name, value):
        self._do_traversal('set', name, value)

    def __delattr__(self, name):
        self._do_traversal('del', name)


class Segment(Element):
    """
    Class representing an HL7 segment.

    :type name: ``basestring``
    :param name: the HL7 name of the segment (e.g. PID)

    :type parent: an instance of :class:`hl7apy.core.Message`, :class:`hl7apy.core.Group` or None
    :param parent: the parent

    :param reference: the reference structure (see :func:`hl7apy.load_reference`)

    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)

    :type validation_level: ``int``
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)

    :type traversal_parent: an instance of :class:`hl7apy.core.Message`, :class:`hl7apy.core.Group` or None
    :param traversal_parent: the temporary parent used during traversal
    """
    child_classes = (Field,)
    child_parser = ('parse_field', 'parse_fields')

    def __init__(self, name=None, parent=None, reference=None, version=None,
                 validation_level=None, traversal_parent=None):

        if name is None:
            raise OperationNotAllowed("Cannot instantiate an unknown Segment")

        super(Segment, self).__init__(name, parent, reference, version,
                                      validation_level, traversal_parent)

        self.cls_attrs.extend(['allow_infinite_children'])
        last_field_structure = self.structure_by_name.items()[-1][1]['ref']
        self.allow_infinite_children = last_field_structure[1] == 'varies'

    def add_field(self, name):
        """
        Create an instance of :class:`hl7apy.core.Field` having the given name

        :param name: the name of the field to be created (e.g. PID_1)
        :return: an instance of :class:`hl7apy.core.Field`

        >>> s = Segment('PID')
        >>> print s.add_field('PID_1')
        <Field PID_1 (SET_ID_PID) of type SI>
        """
        return self.children.create_element(name)

    def find_child_reference(self, name):
        """
        Override the corresponding ``Element``'s method. This because if the Segment allows children
        other then the ones expected in the HL7 structure (i.e. if the last known field is of type `varies`)
        the ``Field`` will be created although it is not found in the HL7 structures, but only if its name is `<SegmentName>_<index>`.
        In this case the method returns a reference for a ``Field`` of type `varies`.

        For example the segment `QPD` in the version 2.5 has the last known child `QPD_3` of type varies
        That means that it allows fields with name QPD_4, QPD_5,...QPD_n.

        """
        name = name.upper()
        element =  self.structure_by_name.get(name, None) or self.structure_by_longname.get(name, None)

        if element is None: # not found in self.structure
            if self.allow_infinite_children and _valid_child_name(name, self.name):
                element = {'cls': Field,
                           'name': name.upper(),
                           'ref': ('leaf', 'varies', None, None)}
            else:
                element = find_reference(name, self.child_classes, self.version)
                if Validator.is_strict(self.validation_level): # cannot be created if validation is strict
                    raise ChildNotValid(name, self)
        return element

    def parse_child(self, text, child_name=None, reference=None):
        kwargs = {'encoding_chars': self.encoding_chars, 'reference': reference, 'name': child_name}
        return super(Segment, self).parse_child(text, **kwargs)

    def parse_children(self, text, **kwargs):
        segment_name = text[:3]
        if segment_name != self.name:
            raise OperationNotAllowed('Cannot assign a segment with a different name')
        text = text[4:] if segment_name != 'MSH' else text[3:]
        kwargs = {'name_prefix': self.name, 'encoding_chars' : self.encoding_chars}
        return super(Segment, self).parse_children(text, **kwargs)

    def to_er7(self, encoding_chars=None, trailing_children=False):
        """
        Return the ER7-encoded string

        :type encoding_chars: ``dict``
        :param encoding_chars: a dictionary containing the encoding chars or None to use the default (see :func:`hl7apy.set_default_encoding_chars`)

        :type trailing_children: ``bool``
        :param trailing_children: if True, trailing children will be added even if their value is None

        :return: the ER7-encoded string

        >>> pid = Segment("PID")
        >>> pid.pid_1 = '1'
        >>> pid.pid_5 = "EVERYMAN^ADAM"
        >>> print pid.to_er7()
        PID|1||||EVERYMAN^ADAM
        """
        if encoding_chars is None:
            encoding_chars = self.encoding_chars

        separator = encoding_chars.get('FIELD')
        repetition = encoding_chars.get('REPETITION')
        s = [self.name]
        for child in self._get_children(trailing_children):
            if child is not None:
                s.append(repetition.join(item.to_er7(encoding_chars) for item in child))
            else:
                try:
                    s.append(self._handle_empty_children(encoding_chars))
                except NotImplementedError:
                    pass

        if self.name == 'MSH' and len(s) > 1:
            s.pop(1)

        return separator.join(s)

    def _is_valid_child(self, child):
        # cannot add an unknown child with strict validation
        if child.name is None and Validator.is_strict(self.validation_level):
            return False
        valid = super(Segment, self)._is_valid_child(child)
        if valid:
            if child.name is not None:
                # cannot add a child with a name that differs from the segment name
                # (e.g. cannot add a PID_1 field to an SPM segment)
                if not child.name.upper().startswith(self.name.upper()):
                    return False
                if self.allow_infinite_children and _valid_child_name(child.name, self.name):
                    return True
        return valid

    def _handle_empty_children(self, encoding_chars=None):
        return ''


class Group(Element):
    """
    Class representing an HL7 segment group

    :type name: ``basestring``
    :param name: the HL7 name of the message (e.g. OML_O33)

    :param reference: the reference structure (see :func:`hl7apy.load_reference`)

    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)

    :type validation_level: ``int``
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)
    """
    child_parser = ('parse_segment', 'parse_segments')

    def __init__(self, name=None, parent=None, reference=None, version=None,
                 validation_level=None, traversal_parent=None):

        self.child_classes = (Segment, Group)
        super(Group, self).__init__(name, parent, reference, version, validation_level, traversal_parent)
        if self.name is None and Validator.is_strict(self.validation_level):
            raise OperationNotAllowed("Cannot instantiate an unknown Element with strict validation")

    def add_segment(self, name):
        """
        Create an instance of :class:`hl7apy.core.Segment` having the given name

        :param name: the name of the segment to be created (e.g. PID)
        :return: an instance of :class:`hl7apy.core.Segment`

        >>> m = Message('QBP_Q11')
        >>> qpd = m.add_segment('QPD')
        >>> print qpd
        <Segment QPD>
        >>> print qpd in m.children
        True
        """
        return self.children.create_element(name)

    def add_group(self, name):
        """
        Create an instance of :class:`hl7apy.core.Group` having the given name

        :param name: the name of the group to be created (e.g. OML_O33_PATIENT)
        :return: an instance of :class:`hl7apy.core.Group`

        >>> m = Message('OML_O33')
        >>> patient = m.add_group('OML_O33_PATIENT')
        >>> print patient
        <Group OML_O33_PATIENT>
        >>> print patient in m.children
        True
        """

        return self.children.create_element(name)

    def parse_child(self, text, child_name=None, reference=None):
        ref = self.find_child_reference(child_name)
        if ref['cls'] == Group:
            g = Group(child_name, validation_level=self.validation_level, version=self.version, reference=ref['ref'])
            g.value = text
            return g
        else:
            kwargs = {'encoding_chars': self.encoding_chars, 'reference': reference}
            return Element.parse_child(self, text, **kwargs)

    def parse_children(self, text, find_groups=True, **kwargs):
        from hl7apy.parser import create_groups

        children = super(Group, self).parse_children(text, **kwargs)
        if self.name and find_groups:
            self.children = []
            create_groups(self, children, validation_level=self.validation_level)
        else:
            self.children = children

    def _set_value(self, value):
        self.parse_children(value)

    def _get_value(self):
        return self.to_er7()

    value = property(_get_value, _set_value)

    def _is_valid_child(self, child):
        if child.name is None and Validator.is_strict(self.validation_level):
            return False
        return Element._is_valid_child(self, child)

    def _get_children(self, trailing=False):
        if Validator.is_strict(self.validation_level):
            children = self.children.get_ordered_children()
        else:
            children = self.children.get_children()
        if not trailing:
            children = _remove_trailing(children)
        return children


class Message(Group):
    """
    Class representing an HL7 message

    :type name: ``basestring``
    :param name: the HL7 name of the message (e.g. OML_O33)

    :param reference: the reference structure (see :func:`hl7apy.load_reference`)

    :type version: ``basestring``
    :param version: the HL7 version (e.g. "2.5"), or ``None`` to use the default (see :func:`hl7apy.set_default_version`)

    :type validation_level: ``int``
    :param validation_level: the validation level. Possible values are those defined in :class:`hl7apy.consts.VALIDATION_LEVEL` class or ``None`` to use the default validation level (see :func:`hl7apy.set_default_validation_level`)

    :type encoding_chars: ``dict``
    :param encoding_chars: a dictionary containing the encoding chars or None to use the default (see :func:`hl7apy.set_default_encoding_chars`)
    """

    def __init__(self, name=None, reference=None, version=None,
                 validation_level=None,
                 encoding_chars=None):

        super(Message, self).__init__(name, None, reference, version,
                                      validation_level)

        if encoding_chars is None:
            encoding_chars = get_default_encoding_chars()

        self.msh = Segment('MSH', version=self.version, validation_level=self.validation_level)
        self.encoding_chars = encoding_chars
        self.msh.msh_7 = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
        self.msh.msh_12 = self.version

    def parse_children(self, text, find_groups=True, **kwargs):
        from hl7apy.parser import get_message_info

        encoding_chars, message_structure, version = get_message_info(text)

        if not self.is_unknown() and self.name != message_structure:
            raise OperationNotAllowed('Cannot assign a message with a different name')
        elif self.is_unknown(): # the message become a known message
            self.name = message_structure
        if self.version != version:
            raise OperationNotAllowed('Cannot assign a message with a different version')
        elif self.encoding_chars != encoding_chars:
            raise OperationNotAllowed('Cannot assign a message with different encoding chars')

        super(Message, self).parse_children(text, find_groups, **kwargs)

    def _get_encoding_chars(self):
        msh_2 = self.msh.msh_2.msh_2_1.children[0].value.value
        return {
            'FIELD' : self.msh.msh_1.msh_1_1.children[0].value.value,
            'COMPONENT' : msh_2[0],
            'REPETITION' : msh_2[1],
            'ESCAPE' : msh_2[2],
            'SUBCOMPONENT' : msh_2[3],
            'GROUP' : '\r',
            'SEGMENT' : '\r',
        }

    def _set_encoding_chars(self, encoding_chars):
        check_encoding_chars(encoding_chars)
        msh_1 = Field('MSH_1')
        msh_2 = Field('MSH_2')
        msh_1.msh_1_1 = encoding_chars['FIELD']
        value = '{0}{1}{2}{3}'.format(encoding_chars['COMPONENT'],
                                      encoding_chars['REPETITION'],
                                      encoding_chars['ESCAPE'],
                                      encoding_chars['SUBCOMPONENT'])
        s = SubComponent(datatype='ST', value=value)
        c = Component(datatype='ST')
        c.add(s)
        msh_2.msh_2_1 = c
        self.msh.msh_1 = msh_1
        self.msh.msh_2 = msh_2

    encoding_chars = property(_get_encoding_chars, _set_encoding_chars)


if __name__ == '__main__':

    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import unittest

import hl7apy
from hl7apy.core import Group, Field, Component, SubComponent
from hl7apy.parser import parse_message, parse_segments, parse_segment, \
                          parse_fields, parse_field, parse_components, \
                          parse_component, parse_subcomponents
from hl7apy.validation import Validator, VALIDATION_LEVEL

class TestValidation(unittest.TestCase):

    def setUp(self):
        hl7apy.set_default_validation_level(VALIDATION_LEVEL.QUIET)
        self.adt_a01 = 'MSH|^~\\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20130101101500||ADT^A01^ADT_A01|0123456789|P|2.5||||AL\r' \
                       'EVN||20130101101500||AAA|AAA|20130101082500\r' \
                       'PID|1||123-456-789^^^HOSPITAL^MR||SURNAME^NAME^A|||M|||1111 SOMEWHERE STREET^^SOMEWHERE^^^USA||555-555-2004~444-333-222|||M\r' \
                       'NK1|1|WOMAN^WIFE|SPO|1111 SOMEWHERE STREET^^SOMEWHERE^^^USA\r' \
                       'PV1|1|I|PATIENT WARD|U||||^REFERRING^DOCTOR^^MD|^CONSULTING^DOCTOR|CAR||||2|A0|||||||||||||||||||||||||||||2013\r' \
                       'IN1|1|INSURANCE PLAN ID^PLAN DESC|COMPANY ID|INSURANCE COMPANY, INC.|5555 INSURERS STREET^^SOMEWHERE^^^USA||||||||||||||||||||||||||||||||||||||||||||555-44-3333\r'

        self.oml_o33 = 'MSH|^~\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20110708162817||OML^O33^OML_O33|978226056138290600|D|2.5|||||USA||EN\r' \
                       'PID|||1010110909194822^^^GATEWAY_IL&1.3.6.1.4.1.21367.2011.2.5.17&ISO^PK||PIPPO^PLUTO^^^^^L||19790515|M|||VIA DI TOPOLINO^CAGLIARI^CAGLIARI^^09100^100^H^^092009^^~^^^^^^L^^^|||||||PPPPPP79E15B354I^^^CF|||||CAGLIARI|||100|||||||||||\r' \
                       'PV1||O|||||||||||||||||1107080001^^^LIS\r' \
                       'SPM|1|100187400201^||SPECIMEN^Blood|||||||PSN^Human Patient||||||20110708162817||20110708162817|||||||1|CONTAINER^CONTAINER DESC\r' \
                       'ORC|NW|83428|83428|18740|SC||||20110708162817||||||||^\r' \
                       'TQ1|||||||||R\r' \
                       'OBR||83428|83428|TPO^ANTI THYROPEROXIDASE ANTIBODIES(TPO)^^TPO||||||||||||ND^UNKNOWN^UNKNOWN\r'

        self.oml_o33_2 = 'MSH|^~\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20110708162817||OML^O33^OML_O33|978226056138290600|D|2.5|||||USA||EN\r' \
                         'PID|||1010110909194822^^^GATEWAY_IL&1.3.6.1.4.1.21367.2011.2.5.17&ISO^PK||PIPPO^PLUTO^^^^^L||19790515|M|||VIA DI TOPOLINO^CAGLIARI^CAGLIARI^^09100^100^H^^092009^^~^^^^^^L^^^|||||||PPPPPP79E15B354I^^^CF|||||CAGLIARI|||100|||||||||||\r' \
                         'PV1||O|||||||||||||||||1107080001^^^LIS\r' \
                         'SPM|1|100187400201^||SPECIMEN^Blood|||||||PSN^Human Patient||||||20110708162817||20110708162817|||||||1|CONTAINER^CONTAINER DESC\r' \
                         'ORC|NW|83428|83428|18740|SC||||20110708162817||||||||^\r' \
                         'TQ1|||||||||R\r' \
                         'OBR||83428|83428|TPO^ANTI THYROPEROXIDASE ANTIBODIES(TPO)^^TPO||||||||||||ND^UNKNOWN^UNKNOWN\r' \
                         'SPM|2|100187400101^||SPECIMEN^Blood|||||||PSN^Human Patient||||||20110708162817||20110708162817|||||||1|CONTAINER^CONTAINER DESC\r' \
                         'ORC|NW|83425|83425|18740|SC||||20110708162817||||||||^\r' \
                         'TQ1|||||||||R\r' \
                         'OBR||83425|83425|CA^S-CALCIUM^^CA||||||||||||ND^UNKNOWN^UNKNOWN\r' \
                         'ORC|NW|83426|83426|18740|SC||||20110708162817||||||||^\r' \
                         'TQ1|||||||||R\r' \
                         'OBR||83426|83426|HDL^HDL CHOLESTEROL^^HDL||||||||||||ND^UNKNOWN^UNKNOWN\r' \
                         'ORC|NW|83427|83427|18740|SC||||20110708162817||||||||^\r' \
                         'TQ1|||||||||R\r' \
                         'OBR||83427|83427|LDL^LDL CHOLESTEROL^^LDL||||||||||||ND^UNKNOWN^UNKNOWN\r'


    def _create_message(self, msg_str):
        return parse_message(msg_str)

    def test_well_structured_message(self):
        """
        Tests that a valid message is validated
        """
        for msg_str in (self.adt_a01, self.oml_o33):
            msg = self._create_message(msg_str)
            self.assertTrue(msg.validate())

    # def test_oml_o33_2_message(self):
    #     # This test is failing because the oml_o33_2 parsing creates a
    #     # OML_O33_ORDER_PRIOR which expect a OBR segment, but the parser's
    #     # implementation puts the OBR segment in the group OML_O33_OBSERVATION_REQUEST
    #     # This behaviour HL7 fault
    #     msg = self._create_message(self.oml_o33_2)
    #     self.assertTrue(msg.validate())

    def test_unknown_message(self):
        msg = self._create_message(self.adt_a01)
        msg.name = None
        self.assertFalse(msg.validate())

    def test_missing_required_group(self):
        """
        Test that if a required group is not present the messase in not validated
        The message misses the oml_o33_specimen group
        """
        msg = self._create_message(self.oml_o33)
        del msg.oml_o33_specimen
        self.assertFalse(msg.validate())
        # print msg.children.indexes.keys()

    def test_missing_required_segment(self):
        """
        Tests that when a required segment is missing the message is not validated
        The message used misses the EVN segment
        """
        msg = self._create_message(self.adt_a01)
        del msg.evn
        self.assertFalse(msg.validate())

    def test_missing_required_field(self):
        """
        Tests that when a required field is missing the message is not validated
        The message used misses the msh_9 segment
        """
        msg = self._create_message(self.adt_a01)
        del msg.msh.msh_9
        self.assertFalse(msg.validate())

    def test_missing_required_component(self):
        """
        Tests that when a required component is missing the message is not validated
        The message used misses the msg_1 component in the msh_9 field
        """
        msg = self._create_message(self.adt_a01)
        del msg.msh.msh_9.msg_1
        self.assertFalse(msg.validate())

    def test_missing_required_subcomponent(self):
        """
        Tests that when a required subcomponent is missing the message is not validated
        The message used misses the cwe_1 subcomponent in the cx_10 component
        """
        msg = self._create_message(self.oml_o33)
        msg.pid.pid_3.cx_10 = 'xxx&yyy&zzz&&&&&&&'
        del msg.pid.pid_3.cx_10.cwe_1
        self.assertFalse(msg.validate())

    def test_having_more_groups(self):
        """
        Tests that when a group occurs more than the allowed times the message is not validated
        The message used has 2 occurrence of oml_o33_group
        """
        msg = self._create_message(self.oml_o33)
        oml_o33_patient = Group('OML_O33_PATIENT')
        segments = parse_segments('PID|||1010110909194822^^^GATEWAY_IL&1.3.6.1.4.1.21367.2011.2.5.17&ISO^PK||PIPPO^PLUTO^^^^^L||19790515|M|||VIA DI TOPOLINO^CAGLIARI^CAGLIARI^^09100^100^H^^092009~^^^^^^L|||||||PPPPPP79E15B354I^^^CF|||||CAGLIARI|||100\rPV1||O|||||||||||||||||1107080001^^^LIS')
        oml_o33_patient.children = segments
        msg.add(oml_o33_patient)
        self.assertFalse(msg.validate())

    def test_having_more_segments(self):
        """
        Tests that when a segment occurs more than the allowed times the message is not validated
        The message used has 2 occurrence of the segment evn
        """
        msg = self._create_message(self.adt_a01)
        evn = parse_segment('EVN||20080115153000||AAA|AAA|20080114003000')
        msg.add(evn)
        self.assertFalse(msg.validate())

    def test_having_more_fields(self):
        """
        Tests that when a field occurs more than the allowed times the message is not validated
        The message used has 2 occurrence of msh_9
        """
        msg = self._create_message(self.adt_a01)
        msh_9 = parse_field('RSP^SLI^RSP_K11', 'MSH_9')
        msg.msh.add(msh_9)
        self.assertFalse(msg.validate())

    def test_wrong_datatype_field(self):
        """
        Tests that if a field is not of the correct datatype the message is not validate
        The message used has the MSH_9 of type ST
        """
        msg = self._create_message(self.adt_a01)
        msh_9 = Field('MSH_9', datatype='ST')
        msh_9.msh_9_1 = 'ADT_A01'
        msg.msh.msh_9 = msh_9
        self.assertFalse(msg.validate())

    def test_wrong_datatype_component(self):
        """
        Tests that if a component is not of the correct datatype the message is not validate
        The message used has the MSG_1 of type ST
        """
        msg = self._create_message(self.adt_a01)
        msg_1 = Component('MSG_1', datatype='ST')
        msg_1.add(SubComponent(datatype='ST', value='ADT_A01'))
        msg.msh.msh_9.msg_1 = msg_1
        self.assertFalse(msg.validate())

    def test_wrong_datatype_subcomponent(self):
        """
        Tests that if a component is not of the correct datatype the message is not validated
        The message used has the CWE_1 of type NM
        """
        msg = self._create_message(self.adt_a01)
        cwe_1 = SubComponent('CWE_1', datatype='NM', value='1')
        msg.pid.pid_3.cx_10.cwe_1 = cwe_1
        self.assertFalse(msg.validate())

    def test_wrong_segment(self):
        """
        Tests that if there is an unexpected segment the message in not validated
        The message used has an unexpected SPM
        """
        msg = self._create_message(self.adt_a01)
        spm = parse_segment('SPM|1|100187400201^||SPECIMEN^Blood|||||||PSN^Human Patient||||||20110708162817||20110708162817|||||||1|CONTAINER^CONTAINER DESC\r')
        msg.add(spm)
        self.assertFalse(msg.validate())

    def test_wrong_group(self):
        """
        Tests that if there is an unexpected segment the message in not validated
        The message used has an unexpected OML_O33_PATIENT
        """
        msg = self._create_message(self.adt_a01)
        oml_o33_patient = Group('OML_O33_PATIENT')
        segments = parse_segments('PID|||1010110909194822^^^GATEWAY_IL&1.3.6.1.4.1.21367.2011.2.5.17&ISO^PK||PIPPO^PLUTO^^^^^L||19790515|M|||VIA DI TOPOLINO^CAGLIARI^CAGLIARI^^09100^100^H^^092009~^^^^^^L|||||||PPPPPP79E15B354I^^^CF|||||CAGLIARI|||100\rPV1||O|||||||||||||||||1107080001^^^LIS')
        oml_o33_patient.children = segments
        msg.add(oml_o33_patient)
        self.assertFalse(msg.validate())

    def test_wrong_field(self):
        """
        Tests that if there is an unexpected field the message in not validated
        The message used has an unexpected unknown field in the pid
        """
        msg = self._create_message(self.adt_a01)
        unkn_field = Field()
        msg.pid.add(unkn_field)
        self.assertFalse(msg.validate())

    def test_wrong_component(self):
        """
        Tests that if there is an unexpected component the message in not validated
        The message used has an unexpected unknown component in the msh_9
        """
        msg = self._create_message(self.adt_a01)
        unkn_component = Component()
        msg.msh.msh_9.add(unkn_component)
        self.assertFalse(msg.validate())

    def test_wrong_subcomponent(self):
        """
        Tests that if there is an unexpected subcomponent the message in not validated
        The message used has an unexpected unknown subcomponent in the cx_10
        """
        msg = self._create_message(self.oml_o33)
        unkn_subcomponent = SubComponent(datatype='ST')
        msg.pid.pid_3.cx_10 = Component('CX_10')
        msg.pid.pid_3.cx_10.add(unkn_subcomponent)
        self.assertFalse(msg.validate())

if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import unittest
from hl7apy.core import Message, Segment, Field, Group, Component, SubComponent, ElementProxy
from hl7apy.exceptions import ChildNotValid, ChildNotFound, OperationNotAllowed, InvalidName, MaxChildLimitReached, \
                              UnsupportedVersion, InvalidEncodingChars

from hl7apy.validation import VALIDATION_LEVEL
from hl7apy.parser import parse_message, parse_segment, parse_field, parse_component

def _get_invalid_encoding_chars():
    return {'COMPONENT' : '$',
            'SUBCOMPONENT' : '@',
            'REPETITION' : 'r',
            'ESCAPE' : '@'}

def _get_test_msg():
    return 'MSH|^~\\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20110708162817||OML^O33^OML_O33|978226056138290600|D|2.5|||||USA||EN\r' \
           'PID|1||566-554-3423^^^GHH^MR||SURNAME^NAME^A|||M|||1111 SOMEWHERE STREET^^SOMEWHERE^^^USA||555-555-2004~444-333-222|||M\r' \
           'PV1||O|||||||||||||||||1107080001^^^LIS\r' \
           'SPM|1|100187400201||SPECIMEN^Blood|||||||PSN^Human Patient||||||20110708162817||20110708162817|||||||1|CONTAINER^CONTAINER DESC\r' \
           'ORC|NW|83428|83428|18740|SC||||20110708162817||||||||\r' \
           'TQ1|||||||||R\r' \
           'OBR||83428|83428|TPO^ANTI THYROPEROXIDASE ANTIBODIES(TPO)^^TPO||||||||||||ND^UNKNOWN^UNKNOWN\r' \
           'SPM|2|100187400101||SPECIMEN^Blood|||||||PSN^Human Patient||||||20110708162817||20110708162817|||||||1|CONTAINER^CONTAINER DESC\r' \
           'ORC|NW|83425|83425|18740|SC||||20110708162817||||||||\rTQ1|||||||||R\r' \
           'OBR||83425|83425|CA^S-CALCIUM^^CA||||||||||||ND^Sconosciuto^Sconosciuto\rORC|NW|83426|83426|18740|SC||||20110708162817||||||||\r' \
           'TQ1|||||||||R\rOBR||83426|83426|HDL^HDL CHOLESTEROL^^HDL||||||||||||ND^UNKNOWN^UNKNOWN\r' \
           'ORC|NW|83427|83427|18740|SC||||20110708162817||||||||\r' \
           'TQ1|||||||||R\rOBR||83427|83427|LDL^LDL CHOLESTEROL^^LDL||||||||||||ND^UNKNOWN^UNKNOWN'


class TestMessage(unittest.TestCase):

    #Message test cases

    def test_create_empty_message(self):
        e = Message()
        self.assertEqual(e.classname, 'Message')
        self.assertRaises(OperationNotAllowed, Message,
                          validation_level=VALIDATION_LEVEL.STRICT)

    def test_create_unknown_message(self):
        self.assertRaises(InvalidName, Message, 'AAA_A01')
        self.assertRaises(InvalidName, Message, 'AAA_A01', validation_level=VALIDATION_LEVEL.STRICT)

    def test_create_unsupported_version_message(self):
        self.assertRaises(UnsupportedVersion, Message, version='2.0')

    def test_create_invalid_encoding_chars_message(self):
        self.assertRaises(InvalidEncodingChars, Message, encoding_chars=_get_invalid_encoding_chars())
        self.assertRaises(InvalidEncodingChars, Message, 'ADT_A01', encoding_chars=_get_invalid_encoding_chars(),
                          validation_level=VALIDATION_LEVEL.STRICT)

    def test_create_insensitive(self):
        e = Message('oml_o35')
        self.assertEqual(e.classname, 'Message')
        self.assertTrue(e.is_named('OML_O35'))

    def test_add_group_to_message(self):
        e = Message('OML_O35')
        self.assertRaises(ChildNotFound, e.add_group, 'UNKNOWN_GROUP')
        g = e.add_group("OML_O35_PATIENT")
        self.assertTrue(g.is_named('OML_O35_PATIENT'))
        self.assertEqual(g.classname, 'Group')

    def test_add_empty_children_to_message(self):
        a = Message('OML_O33', validation_level=VALIDATION_LEVEL.STRICT)
        self.assertRaises(ChildNotValid, a.add, Group())
        #self.assertRaises(ChildNotValid, a.add, Segment())
        b = Message('OML_O33')
        b.add(Group())
        #b.add(Segment())

    def test_add_not_allowed_segment_to_known_message(self):
        a = Message('OML_O33', validation_level=VALIDATION_LEVEL.STRICT)
        self.assertRaises(ChildNotValid, a.add, Segment('MSA'))
        b = Message('OML_O33')
        b.add(Segment('MSA'))

    def test_add_known_segment_to_empty_message(self):
        a = Message('OML_O33')
        a.add(Segment('MSA'))

    def test_add_known_group_to_empty_message(self):
        a = Message('OML_O33')
        a.add(Group('OML_O33_PATIENT'))

    def test_assign_wrong_segment_to_known_position(self):
        a = Message('OML_O33', validation_level=VALIDATION_LEVEL.STRICT)
        b = Message('OML_O33')
        with self.assertRaises(ChildNotValid):
            a.msh = Segment('SPM')
        with self.assertRaises(ChildNotValid):
            a.pid = 'EVN||20080115153000||||20080114003000'
        with self.assertRaises(ChildNotValid):
            b.msh = Segment('SPM')
        with self.assertRaises(ChildNotValid):
            b.pid = 'EVN||20080115153000||||20080114003000'

    def test_add_segment_to_message_mix(self):
        a = Message("OML_O33",  validation_level=VALIDATION_LEVEL.QUIET)
        msh = Segment("MSH", validation_level=VALIDATION_LEVEL.QUIET)
        pid = Segment("PID", validation_level=VALIDATION_LEVEL.QUIET)
        g = Group('OML_O33_PATIENT')
        g.add(pid)
        a.add(msh)
        a.add(g)
        #print a.to_er7()      #check behaviour with when to_er7 is ready
        self.assertTrue(True)

    def test_assign_value(self):
        msg = _get_test_msg()
        a = Message("OML_O33", validation_level=VALIDATION_LEVEL.QUIET)
        parsed_a = parse_message(msg, validation_level=VALIDATION_LEVEL.QUIET)
        a.value = msg
        self.assertEqual(a.to_er7(), parsed_a.to_er7())

        b = Message("OML_O33", validation_level=VALIDATION_LEVEL.STRICT)
        b.value = msg
        parsed_b = parse_message(msg, validation_level=VALIDATION_LEVEL.STRICT)
        self.assertEqual(b.to_er7(), parsed_b.to_er7())
        self.assertEqual(b.children.indexes.keys(), parsed_b.children.indexes.keys())

        c = Message("ADT_A01", validation_level=VALIDATION_LEVEL.QUIET)
        with self.assertRaises(OperationNotAllowed):
            c.value = msg

        msg = msg.replace('^', 'x')
        with self.assertRaises(OperationNotAllowed):
            a.value = msg

        c = Message("OML_O33", version='2.6')
        with self.assertRaises(OperationNotAllowed):
            c.value = msg

    def test_assign_value_unknown_message(self):
        msg = _get_test_msg()
        a = Message()
        parsed_a = parse_message(msg, validation_level=VALIDATION_LEVEL.QUIET)
        a.value = msg
        self.assertEqual(a.name, 'OML_O33')
        self.assertEqual(a.to_er7(), parsed_a.to_er7())

    #def test_message_ordered_children(self):
    #    m = Message('OML_O33')
    #    m.add(Group('OML_O33_PATIENT'))
    #    ordered_children = m.children.get_ordered_children()
    #    self.assertEqual(ordered_children[0][0].name, 'MSH' )
    #    self.assertIsNone(ordered_children[1])
    #    self.assertEqual(ordered_children[3][0].name, 'OML_O33_PATIENT')
    #    self.assertIsNone(ordered_children[2])
    #    self.assertIsNone(ordered_children[4])

    #def test_message_get_children(self):
    #    m = Message('OML_O33')
    #    children = m.children.get_children()
    #    self.assertEqual(len(children), 1)
    #    m.pid = 'PID|||||bianchi^mario|||'
    #    children = m.children.get_children()
    #    self.assertEqual(len(children), 2)


class TestGroup(unittest.TestCase):

    #Group test cases

    def setUp(self):
        self.oml_o33_specimen = 'SPM|1|100187400201||SPECIMEN^Blood|||||||PSN^Human Patient||||||20110708162817||20110708162817|||||||1|CONTAINER^CONTAINER DESC\r' \
           'ORC|NW|83428|83428|18740|SC||||20110708162817||||||||\r' \
           'TQ1|||||||||R\r' \
           'OBR||83428|83428|TPO^ANTI THYROPEROXIDASE ANTIBODIES(TPO)^^TPO||||||||||||ND^UNKNOWN^UNKNOWN\r'

    def test_create_unamed_group_strict(self):
        self.assertRaises(OperationNotAllowed,Group, validation_level=VALIDATION_LEVEL.STRICT)

    def test_add_unexpected_child_to_group(self):
        g = Group()
        m = Message()
        f = Field()
        c = Component(datatype='ST')
        sub = SubComponent(datatype='ST')
        self.assertRaises(ChildNotValid, g.add, m)
        self.assertRaises(ChildNotValid, g.add, f)
        self.assertRaises(ChildNotValid, g.add, c)
        self.assertRaises(ChildNotValid, g.add, sub)

    def test_delete_group(self):
        m = Message('OML_O33',  validation_level=VALIDATION_LEVEL.QUIET)
        g = Group ('OML_O33_PATIENT')
        m.add(g)
        self.assertEqual(m.oml_O33_patient.name, 'OML_O33_PATIENT' )
        del m.oml_o33_patient
        self.assertFalse(g in m.children)

    def test_create_supported_version_group(self):
        Group(version='2.5')

    def test_create_unsupported_version_group(self):
        self.assertRaises(UnsupportedVersion, Group, version='2.0')

    def test_assign_value(self):
        g = Group('OML_O33_SPECIMEN')
        g.value = self.oml_o33_specimen

        g = Group('OML_O33_SPECIMEN', validation_level=VALIDATION_LEVEL.STRICT)
        g.value = self.oml_o33_specimen

    def test_assign_value_traversal(self):
        m1 = Message('OML_O33')
        m2 = Message('OML_O33')
        m1.oml_o33_specimen.value = self.oml_o33_specimen
        m2.oml_o33_specimen = self.oml_o33_specimen
        self.assertEqual(m1.to_er7(), m2.to_er7())


    def test_assign_value_unknown_group(self):
        g = Group()
        g.value = self.oml_o33_specimen


class TestSegment(unittest.TestCase):

    #Segment test cases

    #def test_create_empty_segment(self):
    #    s = Segment()
    #    self.assertEqual(s.classname, 'Segment')
    #    self.assertRaises(OperationNotAllowed, Segment, validation_level=VALIDATION_LEVEL.STRICT)

    def test_create_unknown_segment(self):
        self.assertRaises(InvalidName, Segment, 'XXX')
        self.assertRaises(InvalidName, Segment, 'XXX', validation_level=VALIDATION_LEVEL.STRICT)

    def test_create_empty_segment(self):
        self.assertRaises(OperationNotAllowed, Segment)

    def test_create_unsupported_version_segment(self):
        s = Segment('PID', version = '2.5')
        self.assertRaises(UnsupportedVersion, Segment, 'PID', version='2.0')

    def test_add_field(self):
        e = Segment('PID')
        pid_5 = e.add_field("PID_5")
        self.assertEqual(pid_5.classname, 'Field')
        self.assertTrue(pid_5.is_named('PID_5'))
        self.assertTrue(pid_5.is_named('PATIENT_NAME'))
        self.assertRaises(ChildNotFound, e.add_field, 'UNKNOWN_FIELD')

    def test_traversal_equality(self):
        obr = Segment('OBR')
        obr.obr_26 = 'xxx&yyy^zzz^www'
        obr_26 = obr.parent_result
        self.assertTrue(isinstance(obr_26, ElementProxy))
        self.assertTrue(obr_26[0] == obr.obr_26[0], "obr.parent_result != obr.obr_26")

    def test_traversal_by_name(self):
        obr = Segment('OBR')
        obr_26 = obr.parent_result
        self.assertTrue(obr_26.is_named('OBR_26'))
        self.assertTrue(obr_26.is_named('PARENT_RESULT'))

    def test_recursive_traversal(self):
        obr = Segment('OBR')
        obr.obr_26 = 'xxx&yyy^zzz^www'
        by_name = obr.parent_result.parent_observation_identifier
        by_position = obr.obr_26.obr_26_1
        self.assertEqual(by_name[0], by_position[0]) # bug!

    def test_add_empty_field(self):
        s = Segment('SPM', validation_level=VALIDATION_LEVEL.STRICT)
        self.assertRaises(ChildNotValid, s.add, Field())
        s = Segment('SPM')
        s.add(Field())

    #def test_add_known_fields_to_empty_segment(self):
    #   s = Segment()
    #    #self.assertRaises(ChildNotFound, s.add, Field('spm_10')) #This one is not raised!!!!

    def test_add_not_allowed_fields_to_known_segments(self):
        s = Segment('PID', validation_level=VALIDATION_LEVEL.STRICT)
        self.assertRaises(ChildNotValid, s.add, Field('spm_10'))
        s = Segment('PID')
        self.assertRaises(ChildNotValid, s.add, Field('spm_10'))

    def test_assign_wrong_field_to_known_position(self):
        s1 = Segment('MSH',validation_level=VALIDATION_LEVEL.STRICT)
        s2 = Segment('QPD')
        with self.assertRaises(ChildNotValid):
            s1.msh_10 = Field('spm_10')
        with self.assertRaises(ChildNotValid):
            s2.qpd_3 = Field('pid_3')

    def test_access_to_unknown_field(self):
        s1 = Segment('MSH',validation_level=VALIDATION_LEVEL.STRICT)
        s2 = Segment('PID')
        with self.assertRaises(ChildNotFound):
            s1.msh_100
        with self.assertRaises(ChildNotFound):
            s2.pid_100

    def test_delete_segment(self):
        m = Message('OML_O33')
        pid = Segment('PID')
        m.add(pid)
        del m.pid
        self.assertFalse(pid in m.children)

    def test_assign_value(self):
        segment_str = 'PID|1||123-456-789^^^HOSPITAL^MR||SURNAME^NAME^A|||M|||1111 SOMEWHERE STREET^^SOMEWHERE^^^USA||555-555-2004~444-333-222|||M\r'

        s = Segment('PID')
        parsed_s = parse_segment(segment_str)
        s.value = segment_str
        self.assertEqual(s.to_er7(), parsed_s.to_er7())

        s = Segment('PID', validation_level=VALIDATION_LEVEL.STRICT)
        parsed_a = parse_segment(segment_str)
        s.value = segment_str
        self.assertEqual(s.to_er7(), parsed_a.to_er7())

    def test_assign_value_traversal(self):
        segment_str = 'PID|1||123-456-789^^^HOSPITAL^MR||SURNAME^NAME^A|||M|||1111 SOMEWHERE STREET^^SOMEWHERE^^^USA||555-555-2004~444-333-222|||M\r'

        m1 = Message('OML_O33')
        m2 = Message('OML_O33')
        m1.pid.value = segment_str
        m2.pid.value = segment_str
        self.assertEqual(m1.to_er7(), m2.to_er7())

    def test_assign_wrong_value(self):
        s = Segment('PID')
        wrong_segment_str = 'EVN|1||123-456-789^^^HOSPITAL^MR||SURNAME^NAME^A|||M|||1111 SOMEWHERE STREET^^SOMEWHERE^^^USA||555-555-2004~444-333-222|||M\r'
        with self.assertRaises(OperationNotAllowed):
            s.value = wrong_segment_str


class TestField(unittest.TestCase):

    #Field test cases

    def test_create_empty_field(self):
        f = Field()
        self.assertEqual(f.classname, 'Field')
        self.assertRaises(OperationNotAllowed, Field, validation_level=VALIDATION_LEVEL.STRICT)

    def test_create_unknown_field(self):
        self.assertRaises(InvalidName, Field, 'ckk_10')
        self.assertRaises(InvalidName, Field, 'ckk_10', validation_level=VALIDATION_LEVEL.STRICT)

    def test_create_unsupported_version_field(self):
        # f = Field(version = '2.5')
        self.assertRaises(UnsupportedVersion, Field, version='2.0')

    def test_create_varies_datatype_field(self):
        f = Field ('PID_1', datatype='varies')
        self.assertEqual(f.datatype, 'varies')

    def test_add_empty_component(self):
        f1 = Field('pid_3', validation_level=VALIDATION_LEVEL.STRICT)
        self.assertRaises(ChildNotValid, f1.add, Component(datatype="ST"))
        f2 =  Field('pid_3')
        f2.add(Component(datatype="ST"))

    def test_add_known_components_to_empty_fields(self):
        f1 = Field('pid_3', validation_level=VALIDATION_LEVEL.STRICT)
        self.assertRaises(ChildNotValid, f1.add, Component(datatype='CX_1'))
        f2 = Field('pid_3')
        #self.assertRaises(ChildNotValid, f2.add, Component(datatype='CX_1')) #this one is not raised!!!

    def test_add_not_allowed_components_to_known_field(self):
        f1 = Field('pid_3', validation_level=VALIDATION_LEVEL.STRICT)
        self.assertRaises(ChildNotValid, f1.add, Component(datatype='ST'))
        f2 = Field('pid_3')
        self.assertRaises(ChildNotValid, f2.add, Component('XPN_1'))

    def test_assign_wrong_component_to_known_position(self):
        f1 = Field('pid_10', validation_level=VALIDATION_LEVEL.STRICT)
        with self.assertRaises(ChildNotValid):
            f1.ce_1 = Component('CX_1')
        f2 = Field('pid_3')
        #with self.assertRaises(ChildNotValid):   #this one is not raised!!!
        #	f2.cx_1 = Component('HD_1')
        f2.cx_1 = Component('CX_1')

    def test_access_to_unknown_component(self):
        f1 = Field('pid_10', validation_level=VALIDATION_LEVEL.STRICT)
        f2 = Field('pid_3')
        with self.assertRaises(ChildNotFound):
            f1.ce_100
        with self.assertRaises(ChildNotFound):
            f2.cx_100

    def test_add_more_components_to_base_datatype_field(self):
        f1 = Field('pid_8', validation_level=VALIDATION_LEVEL.STRICT) #this is a base datatype field
        f1.add(Component(datatype='IS'))
        self.assertRaises(MaxChildLimitReached, f1.add, Component(datatype='ST'))
        f2 = Field('pid_8')
        f2.add(Component(datatype='IS'))
        self.assertRaises(MaxChildLimitReached, f2.add, Component(datatype='ST'))

    def test_override_field_datatype_strict(self):
        a = Field('pid_3',  validation_level=VALIDATION_LEVEL.STRICT)
        with self.assertRaises(OperationNotAllowed):
            a.datatype = 'HD'
        self.assertRaises(OperationNotAllowed, Field, 'pid_3', datatype = 'HD', validation_level=VALIDATION_LEVEL.STRICT)

    def test_override_field_datatype(self):
        a = Field('pid_3', 'HD')
        self.assertEqual(a.datatype, 'HD')

    def test_override_field_containing_children_datatype(self):
        a = Field('pid_3')
        a.cx_1 = 'cx_1 value'
        a.cx_4 = 'cx_4 value'
        with self.assertRaises(OperationNotAllowed):
            a.datatype = 'HD'

    def test_delete_field(self):
        m = Message('OML_O33')
        msh = m.msh
        msh7 = msh.msh_7
        del msh.msh_7
        self.assertTrue(msh7 not in msh.children)

    #def test_create_fields_by_get(self):
    #    s = Segment('MSH')
    #    s.msh_18

    def test_base_datatype_field_get(self):
        s = Segment('MSH')
        s.msh_10.msh_10_1
        s.msh_10.msh_10_1.value = 'Value'
        s.msh_10.msh_10_1.value = '11111'
        with self.assertRaises(ChildNotFound):
            s.msh_10.msh_10_2

    def test_assign_field_by_get(self):
        s = Segment('MSH')
        s.msh_10 = 'Value'
        s.msh_10.msh_10_1 = 'Value'

    def test_add_component(self):
        f = Field('PID_5')
        f.add_component('XPN_1')
        self.assertEqual(f.children[0].name,'XPN_1')

    def test_assign_value(self):
        field_str = '1010110909194822^^^AUTH&1.3.6.1.4.1.21367.2011.2.5.17&ISO^PK'

        f = Field('PID_3')
        f.value = field_str
        parsed_field = parse_field(field_str, 'PID_3')
        self.assertEqual(f.to_er7(), parsed_field.to_er7())

        f = Field('PID_3', validation_level=VALIDATION_LEVEL.STRICT)
        f.value = field_str
        self.assertEqual(f.to_er7(), parsed_field.to_er7())

        f = Field('PID_1', validation_level=VALIDATION_LEVEL.STRICT)
        with self.assertRaises(MaxChildLimitReached):
            f.value = '1^2'

    def test_assign_value_traversal(self):
        field_str = '1010110909194822^^^AUTH&1.3.6.1.4.1.21367.2011.2.5.17&ISO^PK'

        s1 = Segment('PID')
        s2 = Segment('PID')
        s1.pid_3.value = field_str
        s2.pid_3 = field_str

        self.assertEqual(s1.to_er7(), s2.to_er7())

    def test_assign_value_unknown_field(self):
        field_str = '1010110909194822^^^AUTH&1.3.6.1.4.1.21367.2011.2.5.17&ISO^PK'
        f = Field()
        f.value = field_str

    def test_assign_value_with_field_separator(self):
        field_str = 'xxx|yyy'
        escaped_str = 'xxx\F\yyy'
        f = Field('PID_3')
        f.value = field_str
        self.assertEqual(f.to_er7(), escaped_str)

        f = Field('PID_3', validation_level=VALIDATION_LEVEL.STRICT)
        f.value = field_str
        self.assertEqual(f.to_er7(), escaped_str)

        f = Field()
        f.value = field_str
        self.assertEqual(f.to_er7(), 'xxx\F\yyy')

    def test_assign_value_with_repetition(self):
        """
        Check that the repetition encoding chars is handled as a normal char, thus escaped when converting the field to er7
        :return:
        """
        field_str = 'xxx~yyy'
        f = Field()
        f.value = field_str
        self.assertEqual(f.to_er7(), 'xxx\R\yyy')

        f = Field('PID_2')
        f.value = field_str
        self.assertEqual(f.to_er7(), 'xxx\R\yyy')

        f = Field('PID_2', validation_level=VALIDATION_LEVEL.STRICT)
        f.value = field_str
        self.assertEqual(f.to_er7(), 'xxx\R\yyy')


class TestComponent(unittest.TestCase):

#Component test cases

    def test_create_empty_component_strict(self):
        c = Component(datatype="ST")
        self.assertEqual(c.classname, 'Component')
        c1 = Component(datatype="ST", validation_level=VALIDATION_LEVEL.STRICT)

    def test_create_empty_component(self):
        c = Component()
        self.assertEqual(c.classname, 'Component')

    def test_create_empty_component_strict(self):
        self.assertRaises(OperationNotAllowed, Component, validation_level=VALIDATION_LEVEL.STRICT)

    def test_create_unknown_component(self):
        self.assertRaises(InvalidName, Component, 'xxx_1')
        self.assertRaises(InvalidName, Component, 'xxx_1', validation_level=VALIDATION_LEVEL.STRICT)

    def test_create_invalid_component(self):
        self.assertRaises(InvalidName, Component, 'AD')

    def test_create_unsupported_version_component(self):
        c = Component(datatype = 'ST', version = '2.5')

    def test_add_empty_subcomponent(self):
        c1 = Component('cx_4', validation_level=VALIDATION_LEVEL.STRICT)
        #self.assertRaises(ChildNotValid, c1.add, SubComponent()) #Strange Exception raised here: expected string or buffer(*)
        c2 = Component('cx_4')
        #c2.add(SubComponent()) #the same as (*)

    def add_unknown_component_strict(self):
        self.assertRaises(OperationNotAllowed, Component, validation_level=VALIDATION_LEVEL.STRICT)

    def test_add_known_subcomponent_to_empty_component(self):
        c = Component()
        self.assertRaises(ChildNotValid, c.add, SubComponent('fn_1'))

    def test_add_not_allowed_subcomponent_to_known_component(self):
        c = Component('cx_4', validation_level=VALIDATION_LEVEL.STRICT)
        self.assertRaises(ChildNotValid, c.add, SubComponent('fn_1'))
        c1 = Component('cx_4')
        self.assertRaises(ChildNotValid, c1.add, SubComponent('fn_1'))

    def test_assign_wrong_subcomponent_to_known_position(self):
        c = Component('XPN_1')
        with self.assertRaises(ChildNotValid):
            c.fn_1 = SubComponent('hd_1')
        c1 = Component('XPN_1', validation_level=VALIDATION_LEVEL.STRICT)
        with self.assertRaises(ChildNotValid):
            c1.fn_1 = SubComponent('hd_1')

    def test_access_to_unknown_subcomponent(self):
        c1 = Component('XPN_1', validation_level=VALIDATION_LEVEL.STRICT)
        c2 = Component('cx_4')
        with self.assertRaises(ChildNotFound):
            c1.fn_100
        with self.assertRaises(ChildNotFound):
            c2.hd_100

    def test_add_more_subcomponents_to_base_datatype_component(self):
        c = Component(datatype='ST')
        c.add(SubComponent(datatype='ST'))
        self.assertRaises(MaxChildLimitReached, c.add, SubComponent(datatype="ST"))
        #c1 = Component(datatype='ST', validation_level=VALIDATION_LEVEL.STRICT)
        #self.assertRaises(ChildNotValid, c1.add, SubComponent(datatype="ST"))

    def test_override_datatype_strict(self):
        c = Component('CX_1',  validation_level=VALIDATION_LEVEL.STRICT)
        with self.assertRaises(OperationNotAllowed):
            c.datatype = 'TX'
        c1 = Component('CX_1', datatype = 'ST', validation_level=VALIDATION_LEVEL.STRICT)
        with self.assertRaises(OperationNotAllowed):
            c1.datatype = 'TX'

    def test_override_valued_datatype(self):
        c = Component('CX_1')
        c.datatype = 'TX'
        self.assertEqual(c.datatype, 'TX')

    #def test_override_none_datatype(self):
    #    c = Component('CX_1')
    #    with self.assertRaises(OperationNotAllowed):
    #        c.datatype = 'TX'

    # def test_add_subcomponent_to_unknown_component(self):
    #     c = Component()
    #     self.assertRaises(ChildNotValid, c.add, SubComponent(datatype = 'ST'))

    def test_add_unexpected_child_to_component(self):
        g = Group()
        m = Message()
        f = Field()
        c_base = Component(datatype="ST")
        c_complex = Component(datatype="CWE")
        sub = SubComponent(datatype="ST")
        #Is this the correct Exception? or shall it be ChildNotValid?
        self.assertRaises(ChildNotValid, c_base.add, g)
        self.assertRaises(ChildNotValid, c_base.add, m)
        self.assertRaises(ChildNotValid, c_base.add, f)
        self.assertRaises(ChildNotValid, c_complex.add, g)
        self.assertRaises(ChildNotValid, c_complex.add, m)
        self.assertRaises(ChildNotValid, c_complex.add, f)

    def test_delete_component(self):
        m = Message('OML_O33')
        m.pid = 'PID|||||bianchi^mario|||'
        pid51 = m.pid.pid_5.pid_5_1
        del m.pid.pid_5.pid_5_1

    def test_create_component_by_get(self):
        s = Segment('MSH')
        #no children are created here
        with self.assertRaises(IndexError):
            s.msh_9.msh_9_1
            s.children[0].children[0].name

    def test_assign_complex_field_datatype_by_get(self):
        p = Segment('PID')
        p.pid_5 = 'test^test'
        self.assertEqual(p.pid_5.xpn_1.to_er7(), 'test')
        self.assertEqual(p.pid_5.xpn_2.to_er7(), 'test')

    def test_create_base_datatype_component_by_get(self):
        f = Field('SID_2')
        f.sid_2_1 = 'field'
        self.assertEqual(f.sid_2_1.to_er7(), 'field')

    def test_add_subcomponent_to_unknown_base_dt_component(self):
        c = Component(datatype='ST')
        self.assertRaises(ChildNotValid, c.add_subcomponent, 'HD_1')

    def test_add_subcomponent(self):
        c = Component('XPN_1')
        c.add_subcomponent('FN_1')
        self.assertEqual(c.children[0].name, 'FN_1')

    def test_assign_value(self):
        cmp_str = 'xxx'
        c = Component('CWE_1')
        c.value = cmp_str
        parsed_cmp = parse_component(cmp_str, 'CWE_1')
        self.assertEqual(c.to_er7(), parsed_cmp.to_er7())

        c = Component('CWE_1', validation_level=VALIDATION_LEVEL.STRICT)
        with self.assertRaises(MaxChildLimitReached):
            c.value = '1&2'

        complex_cmp_str = 'xxx&yyy&zzz'
        c = Component('CX_10', validation_level=VALIDATION_LEVEL.STRICT)
        c.value = complex_cmp_str
        parsed_cmp = parse_component(complex_cmp_str, 'CX_10', datatype='CWE', validation_level=VALIDATION_LEVEL.STRICT)
        self.assertEqual(c.to_er7(), parsed_cmp.to_er7())

    def test_assign_value_traversal(self):
        cmp_str = 'xxx'
        f1 = Field('PID_39')
        f2 = Field('PID_39')
        f1.cwe_1.value = cmp_str
        f2.cwe_1 = cmp_str
        self.assertEqual(f1.to_er7(), f2.to_er7())

        s1 = Segment('PID')
        s2 = Segment('PID')
        s1.pid_39.pid_39_1.value = cmp_str
        s2.pid_39.pid_39_1 = cmp_str

        complex_cmp_str = 'xxx&yyy&zzz'
        f1 = Field('PID_4')
        f2 = Field('PID_4')
        f1.cx_10.value = complex_cmp_str
        f2.cx_10 = complex_cmp_str
        self.assertEqual(f1.to_er7(), f2.to_er7())

        s1.pid_4.pid_4_1.value = complex_cmp_str
        s2.pid_4.pid_4_1 = complex_cmp_str
        self.assertEqual(f1.to_er7(), f2.to_er7())


    def test_assign_value_unknown_component(self):
        cmp_str = 'xxx'
        complex_cmp_str = 'xxx&&&&yyy'
        c = Component()
        c.value = cmp_str
        c.value = complex_cmp_str

    def test_assign_value_with_component_separator(self):
        cmp_str = 'xxx^yyy'
        c = Component()
        c.value = cmp_str
        self.assertEqual(c.to_er7(), 'xxx\S\yyy')

        c = Component('CWE_1')
        c.value = cmp_str
        self.assertEqual(c.to_er7(), 'xxx\S\yyy')

        c = Component('CWE_1', validation_level=VALIDATION_LEVEL.STRICT)
        c.value = cmp_str
        self.assertEqual(c.to_er7(), 'xxx\S\yyy')


class TestSubComponent(unittest.TestCase):

     #SubComponent test cases

    def test_create_subcomponent(self):
        a = SubComponent('HD_1', datatype = 'ST')
        self.assertEqual(a.classname, 'SubComponent')
        b = SubComponent(datatype = 'ST')
        self.assertEqual(b.classname, 'SubComponent')

    def test_create_invalid_subcomponent_empty(self):
        self.assertRaises(OperationNotAllowed, SubComponent)

    def test_create_invalid_subcomponent_with_complex_component_name(self):
        self.assertRaises(OperationNotAllowed, SubComponent, 'CX_4', datatype='ST')

    def test_create_unsupported_version_subcomponent(self):
        s = SubComponent(datatype = 'ST', version = '2.5')
        self.assertRaises(UnsupportedVersion, SubComponent, datatype = 'ST', version = '2.0')

    def test_change_datatype_strict(self):
        self.assertRaises(OperationNotAllowed, SubComponent, 'HD_1', datatype='TX', validation_level=VALIDATION_LEVEL.STRICT)

    def test_change_datatype_for_valued_subcomponent(self):
        a = SubComponent('HD_1', value='value')
        with self.assertRaises(OperationNotAllowed):
            a.datatype = 'ST'

    def test_change_datatype_subcomponent_strict(self):
        s = SubComponent (datatype='ST', validation_level=VALIDATION_LEVEL.STRICT)
        with self.assertRaises(OperationNotAllowed):
            s.datatype = 'TX'

    def test_assign_not_allowed_datatype_to_subcomponent(self):
        a = SubComponent('HD_1')
        with self.assertRaises(OperationNotAllowed):
            a.datatype = 'CX'

    def test_assign_value_traversal(self):
        subcmp_str = 'xxx'

        c1 = Component('CX_10')
        c2 = Component('CX_10')
        c1.cwe_1 = subcmp_str
        c2.cwe_1.value = subcmp_str
        # c2.cwe_1 = subcmp_str
        self.assertEqual(c1.to_er7(), c2.to_er7())
        #
        # s1 = Segment('PID')
        # s2 = Segment('PID')
        # s1.pid_4.pid_4_10_1.value = subcmp_str
        # s2.pid_4.pid_4_10_1 = subcmp_str
        # self.assertEqual(s1.to_er7(), s2.to_er7())
        # complex_cmp_str = 'xxx&yyy&zzz'
        # f1 = Field('PID_4')
        # f2 = Field('PID_4')
        # f1.cx_10.value = complex_cmp_str
        # f2.cx_10 = complex_cmp_str
        # self.assertEqual(f1.to_er7(), f2.to_er7())
        #
        # s1.pid_4.pid_4_1.value = complex_cmp_str
        # s2.pid_4.pid_4_1 = complex_cmp_str
        # self.assertEqual(f1.to_er7(), f2.to_er7())

    def test_add_child_to_subcomponent(self):
        a = SubComponent('HD_1')
        self.assertRaises(OperationNotAllowed, a.add, SubComponent('HD_2'))

    def test_create_subcomponent_by_get(self):
        p = Segment('PID')
        self.assertEqual(p.pid_5.pid_5_1.fn_1.name, 'FN_1')

    def test_create_base_datatype_subcomponent_by_get(self):
        f = Field('STF_2')
        f.stf_2_10 = 'subcomponent'

        #p = Segment('PID')
        #p.pid_3.cx_4.hd_1.value = 'value'
        #p.pid_3.cx_4.hd_1 = 'value' #this raises exception

    def create_unknown_subcomponent_by_get(self):
        f = Field('STF_2')
        with self.assertRaises(ChildNotFound):
            f.stf_2_10_100 = 'subcomponent'






if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import unittest

from hl7apy.parser import parse_message, parse_segments, parse_segment, parse_fields, parse_field, parse_components, parse_component, parse_subcomponent, parse_subcomponents
from hl7apy.validation import VALIDATION_LEVEL
from hl7apy.exceptions import ParserError, OperationNotAllowed, InvalidEncodingChars, InvalidName

class TestParser(unittest.TestCase):

    def _get_multiple_segments_groups_message(self):

        msg_o33 = 'MSH|^~\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20110708162817||OML^O33^OML_O33|978226056138290600|D|2.5|||||USA||EN\r' \
                  'PID|||1010110909194822^^^GATEWAY_IL&1.3.6.1.4.1.21367.2011.2.5.17&ISO^PK||PIPPO^PLUTO^^^^^L||19790515|M|||VIA DI TOPOLINO^CAGLIARI^CAGLIARI^^09100^100^H^^092009^^~^^^^^^L^^^|||||||PPPPPP79E15B354I^^^CF|||||CAGLIARI|||100|||||||||||\r' \
                  'PV1||O|||||||||||||||||1107080001^^^LIS\r' \
                  'SPM|1|100187400201^||SPECIMEN^Blood|||||||PSN^Human Patient||||||20110708162817||20110708162817|||||||1|CONTAINER^CONTAINER DESC\r' \
                  'ORC|NW|83428|83428|18740|SC||||20110708162817||||||||^\r' \
                  'TQ1|||||||||R\r' \
                  'OBR||83428|83428|TPO^ANTI THYROPEROXIDASE ANTIBODIES(TPO)^^TPO||||||||||||ND^UNKNOWN^UNKNOWN\r' \
                  'SPM|2|100187400101^||SPECIMEN^Blood|||||||PSN^Human Patient||||||20110708162817||20110708162817|||||||1|CONTAINER^CONTAINER DESC\r' \
                  'ORC|NW|83425|83425|18740|SC||||20110708162817||||||||^\r' \
                  'TQ1|||||||||R\r' \
                  'OBR||83425|83425|CA^S-CALCIUM^^CA||||||||||||ND^UNKNOWN^UNKNOWN\r' \
                  'ORC|NW|83426|83426|18740|SC||||20110708162817||||||||^\r' \
                  'TQ1|||||||||R\r' \
                  'OBR||83426|83426|HDL^HDL CHOLESTEROL^^HDL||||||||||||ND^UNKNOWN^UNKNOWN\r' \
                  'ORC|NW|83427|83427|18740|SC||||20110708162817||||||||^\r' \
                  'TQ1|||||||||R\r' \
                  'OBR||83427|83427|LDL^LDL CHOLESTEROL^^LDL||||||||||||ND^UNKNOWN^UNKNOWN\r'

        return msg_o33

    def _get_invalid_encoding_chars(self):
        return {'COMPONENT' : '$',
                'SUBCOMPONENT' : '@',
                'REPETITION' : 'r',
                'ESCAPE' : '@'}

    def _get_custom_encoding_chars(self):
        return {'SEGMENT' : '\r',
                'FIELD' :  '@',
                'COMPONENT' : '$',
                'SUBCOMPONENT' : '%',
                'REPETITION' : '*',
                'ESCAPE' : '\\'}

    def test_parse_message_ignoring_groups(self):
        msh = "MSH|^~\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20080115153000||ADT^A01^ADT_A01|0123456789|P|2.5||||AL\r"
        evn = "EVN||20080115153000||AAA|AAA|20080114003000\r"
        pid = "PID|1||123-456-789^^^HOSPITAL^MR||SURNAME^NAME^A|||M|||1111 SOMEWHERE STREET^^SOMEWHERE^^^USA||555-555-2004~444-333-222|||M\r"
        nk1 = "NK1|1|WOMAN^WIFE|SPO|1111 SOMEWHERE STREET^^SOMEWHERE^^^USA\r"
        pv1 = "PV1|1|I|PATIENT WARD|U||||^REFERRING^DOCTOR^^MD|^CONSULTING^DOCTOR|CAR||||2|A0|||||||||||||||||||||||||||||2013\r"
        in1 = "IN1|1|INSURANCE PLAN ID^PLAN DESC|COMPANY ID|INSURANCE COMPANY, INC.|5555 INSURERS STREET^^SOMEWHERE^^^USA||||||||||||||||||||||||||||||||||||||||||||555-44-3333"

        str_message = msh+evn+pid+nk1+pv1+in1
        message = parse_message(str_message, find_groups = False)

        self.assertEqual(message.children[0].name, 'MSH')
        self.assertEqual(message.children[1].name, 'EVN')
        self.assertEqual(message.children[2].name, 'PID')
        self.assertEqual(message.children[3].name, 'NK1')
        self.assertEqual(message.children[4].name, 'PV1')
        self.assertEqual(message.children[5].name, 'IN1')
        self.assertEqual(message.pid.pid_5.pid_5_1.fn_1.to_er7(), 'SURNAME')
        self.assertEqual(message.evn.evn_5.xcn_1.to_er7(), message.evn.evn_5.evn_5_1.to_er7())
        m_string= message.to_er7()
        self.assertEqual(m_string, str_message)

    def test_parse_message_create_groups(self):
        msg = self._get_multiple_segments_groups_message()
        message = parse_message(msg)
        self.assertEqual(message.children[0].name, 'MSH')
        self.assertEqual(message.children[1].name, 'OML_O33_PATIENT')
        self.assertEqual(message.children[2].name, 'OML_O33_SPECIMEN')
        self.assertEqual(message.children[3].name, 'OML_O33_SPECIMEN')
        #the first specimen has only one order
        specimen_1 = message.children[2]
        self.assertEqual(specimen_1.children[0].name, 'SPM')
        self.assertEqual(specimen_1.children[1].name, 'OML_O33_ORDER')
        #the second specimen has three different orders, and the second and third are inside the "ORDER PRIOR" group"
        specimen_2 = message.children[3]
        self.assertEqual(specimen_2.children[1].name, 'OML_O33_ORDER')
        spm_2_order =  specimen_2.children[1]
        self.assertEqual(spm_2_order.children[0].name, 'ORC')
        self.assertEqual(spm_2_order.children[1].name, 'OML_O33_TIMING')
        spm_2_obs_req = spm_2_order.oml_o33_observation_request
        self.assertEqual(spm_2_obs_req.children[0].name, 'OBR')
        self.assertEqual(spm_2_obs_req.children[1].name, 'OML_O33_PRIOR_RESULT')

    def test_message_inspection_by_group_parsing(self):
        msg = self._get_multiple_segments_groups_message()
        message = parse_message(msg)
        #access first spm_27_1: P2_S1
        self.assertEqual(message.oml_o33_specimen[0].spm.spm_27.spm_27_1.to_er7(), 'CONTAINER')
        #by datatype name instead that by position it is the same than previous one
        self.assertEqual(message.oml_o33_specimen[0].spm.spm_27.cwe_1.to_er7(), 'CONTAINER')
        #access to the second order test belonging to the second specimen
        self.assertEqual(message.oml_o33_specimen[
                             1].oml_o33_order.oml_o33_observation_request.oml_o33_prior_result.oml_o33_order_prior[
                             1].obr.obr_4.obr_4_1.to_er7(), 'HDL')
        #try to access to non-existent field
        # with self.assertRaises(IndexError): # IndexError cannot be raised here to maintain lazy instantiation of children during traversal
        #     message.oml_o33_specimen[1].oml_o33_order.oml_o33_observation_request.oml_o33_prior_result.oml_o33_order_prior[2].obr.obr_4.obr_4_1.to_er7()
        #self.assertEqual(message.to_er7(), msg)


    def test_parse_message_with_repetition_segments_and_groups(self):
        msg = self._get_multiple_segments_groups_message()
        m = parse_message(msg)

    def test_parse_invalid_message(self):
        msh = "PID|^~\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20080115153000||ADT^A01^ADT_A01|0123456789|P|2.6||||AL\r"
        pid = "PID|1||123-456-789^^^HOSPITAL^MR||SURNAME^NAME^A|||M|||1111 SOMEWHERE STREET^^SOMEWHERE^^^USA||555-555-2004~444-333-222|||M"
        msg = msh  + pid
        self.assertRaises(ParserError, parse_message, msg)

    def test_parse_message_missing_encoding_chars_zero(self):
        a = 'MSH|||||||||||||||||||'
        self.assertRaises(InvalidEncodingChars, parse_message, a)
        self.assertRaises(InvalidEncodingChars, parse_message, a, validation_level=VALIDATION_LEVEL.STRICT)

    def test_parse_message_missing_encoding_chars(self):
        a = 'MSH|@%\|||||||||'
        self.assertRaises(InvalidEncodingChars, parse_message, a)
        self.assertRaises(InvalidEncodingChars, parse_message, a, validation_level=VALIDATION_LEVEL.STRICT)

    def test_parse_message_too_many_encoding_chars(self):
        a = 'MSH|@%\&$?!|||||||||'
        self.assertRaises(InvalidEncodingChars, parse_message, a)
        self.assertRaises(InvalidEncodingChars, parse_message, a, validation_level=VALIDATION_LEVEL.STRICT)

    def test_parse_unknown_message_strict(self):
        message = "MSH|^~\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20080115153000||ADT^A01^ADT_A01|0123456789|P|2.6||||AL\r"
        parse_message(message, validation_level=VALIDATION_LEVEL.STRICT)
        message = "MSH|^~\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20080115153000|||0123456789|P|2.6||||AL\r"
        self.assertRaises(OperationNotAllowed, parse_message, message, validation_level=VALIDATION_LEVEL.STRICT)

    def test_parse_message_duplicate_encoding_chars(self):
        message = "MSH|^&&^|SENDING APP|SENDING FAC|REC APP|REC FAC|20080115153000||ADT^A01^ADT_A01|0123456789|P|2.6||||AL\r"
        self.assertRaises(InvalidEncodingChars, parse_message, message)
        self.assertRaises(InvalidEncodingChars, parse_message, message, validation_level=VALIDATION_LEVEL.STRICT)

    def test_parse_message_missing_structure(self):
        msh = "MSH|^~\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20080115153000|||0123456789|P|2.6||||AL\r"
        pid = "PID|1||123-456-789^^^HOSPITAL^MR||SURNAME^NAME^A|||M|||1111 SOMEWHERE STREET^^SOMEWHERE^^^USA||555-555-2004~444-333-222|||M\r"
        msg = msh+pid
        parse_message(msg)

    def test_parse_message_missing_structure_strict(self):
        msh = "MSH|^~\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20080115153000|||0123456789|P|2.6||||AL\r"
        pid = "PID|1||123-456-789^^^HOSPITAL^MR||SURNAME^NAME^A|||M|||1111 SOMEWHERE STREET^^SOMEWHERE^^^USA||555-555-2004~444-333-222|||M\r"
        msg = msh+pid
        self.assertRaises(OperationNotAllowed, parse_message, msg, validation_level = VALIDATION_LEVEL.STRICT)

    def test_parse_message_incomplete_structure(self):
        m = parse_message('MSH|^~\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20080115153000||ADT^|')

    def test_parse_message_missing_version(self):
        m = parse_message('MSH|^~\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20080115153000||ADT^A01^ADT_A01||')


    def test_parse_segments(self):
        msh = "MSH|^~\&|SENDING APP|SENDING FAC|REC APP|REC FAC|20080115153000||ADT^A01^ADT_A01|0123456789|P|2.5||||AL\r"
        pid = "PID|1||123-456-789^^^HOSPITAL^MR||SURNAME^NAME^A|||M|||1111 SOMEWHERE STREET^^SOMEWHERE^^^USA||555-555-2004~444-333-222|||M"

        segments_str = msh+pid
        segments = parse_segments(segments_str)
        self.assertEqual(msh.replace('\r',''), segments[0].to_er7())
        self.assertEqual(pid, segments[1].to_er7())

    def test_parse_segment(self):
        segment = "PV1|1|I|PATIENT WARD|U||||^REFERRING^DOCTOR|^CONSULTING^DOCTOR|CAR||||2|A0|||||||||||||||||||||||||||||2008"
        pv1 = parse_segment(segment)
        self.assertEqual(pv1.name, 'PV1')
        self.assertEqual(pv1.pv1_2.to_er7(), 'I')

    def test_parse_segment_invalid_encoding_chars(self):
        segment = 'EVN@@20080115153000@@@@20080114003000\r'
        self.assertRaises(InvalidEncodingChars, parse_segment, segment, encoding_chars = self._get_invalid_encoding_chars())
        self.assertRaises(InvalidEncodingChars, parse_segment, segment, encoding_chars = self._get_invalid_encoding_chars(), validation_level = VALIDATION_LEVEL.STRICT)

    def test_parse_segment_custom_encoding_chars(self):
        segment = "PV1@1@I@PATIENT WARD@U@@@@$REFERRING$DOCTOR$$MD@$CONSULTING$DOCTOR$P@CAR@@@@2@A0@@@@@@@@"
        s = parse_segment(segment, encoding_chars = self._get_custom_encoding_chars())
        self.assertEqual(s.pv1_3.to_er7(), 'PATIENT WARD')
        self.assertEqual(s.pv1_8.pv1_8_2.to_er7(), 'REFERRING')

    def test_parse_segment_all_values_null(self):
        segment = 'PID|||||||||||'
        s = parse_segment(segment)
        self.assertEqual(len(s.children), 0)
        s.pid_2.to_er7()
        s.pid_23.to_er7()
        self.assertEqual(len(s.children), 0)
        s.pid_23 = ''
        self.assertEqual(s.to_er7(), 'PID|||||||||||||||||||||||')

    def test_parse_segment_containing_multiple_instances_field(self):
        segment = 'QPD|||@PID.3.1^aaaa~@PID.8^F'
        s = parse_segment(segment)
        self.assertEqual(len(s.qpd_3), 2)
        #check single PID_3 instances
        self.assertEqual(s.qpd_3[0].qpd_3_1.to_er7(), '@PID.3.1' )
        self.assertEqual(s.qpd_3[0].qpd_3_2.to_er7(), 'aaaa')
        self.assertEqual(s.qpd_3[1].qpd_3_1.to_er7(), '@PID.8' )
        self.assertEqual(s.qpd_3[1].qpd_3_2.to_er7(), 'F')

    def test_parse_fields(self):
        #note: this method seems not to work without segment definition, and also for MSH segment
        field = 'PID|xxx|yyy^zz'
        fields = parse_fields(field)
        for field in fields:
            self.assertEqual(field.classname, 'Field')
        self.assertEqual(fields[0].to_er7(), 'PID')
        self.assertEqual(fields[1].to_er7(), 'xxx')

    def test_parse_fields_multiple_instances(self):
        field = '||@PID.3.1^aaaa^~@PID.8^F'
        f = parse_fields(field, name_prefix = 'PID')
        self.assertEqual(len(f), 2) #only the valued field is recognized
        self.assertEqual(f[0].cx_1.to_er7(), '@PID.3.1')

    def test_parse_null_fields(self):
        fields = 'PID|||'
        f = parse_fields(fields)
        self.assertEqual(len(f), 4)

    def test_parse_fields_invalid_encoding_chars(self):
        fields = 'PID|xxx|yyy^zz'
        self.assertRaises(InvalidEncodingChars, parse_fields, fields, encoding_chars = self._get_invalid_encoding_chars())
        self.assertRaises(InvalidEncodingChars, parse_fields, fields, encoding_chars = self._get_invalid_encoding_chars(), validation_level = VALIDATION_LEVEL.STRICT)

    def test_parse_fields_custom_encoding_chars(self):
        fields = 'xxx@yyy$zz'
        f = parse_fields(fields, encoding_chars = self._get_custom_encoding_chars())
        self.assertEqual(f[0].to_er7(encoding_chars=self._get_custom_encoding_chars()), 'xxx')
        self.assertEqual(f[1].children[0].to_er7(), 'yyy')

    def test_parse_field(self):
        field = 'xxx^yyy^zzz'
        f = parse_field(field) #exception thrown here

    def test_parse_field_invalid_encoding_chars(self):
        field = 'xxx^yyy^zzz'
        self.assertRaises(InvalidEncodingChars, parse_field, field, encoding_chars = self._get_invalid_encoding_chars())
        self.assertRaises(InvalidEncodingChars, parse_field, field, encoding_chars = self._get_invalid_encoding_chars(), validation_level = VALIDATION_LEVEL.STRICT)

    def test_parse_field_custom_encoding_chars(self):
        field = 'aaa$bbb$ccc'
        f = parse_field(field, encoding_chars=self._get_custom_encoding_chars())
        self.assertEqual(len(f.children), 3)
        self.assertEqual(f.children[0].to_er7(), 'aaa')
        self.assertEqual(f.children[1].to_er7(), 'bbb')
        self.assertEqual(f.children[2].to_er7(), 'ccc')

        field = 'aaa^bbb^ccc'
        to_string_field = 'aaa\\S\\bbb\\S\\ccc'
        f =  parse_field(field, encoding_chars=self._get_custom_encoding_chars())
        self.assertEqual(len(f.children), 1)
        self.assertEqual(f.children[0].to_er7(), to_string_field)
        self.assertEqual(f.children[0].to_er7(encoding_chars=self._get_custom_encoding_chars()), field)

    def test_parse_invalid_name_field(self):
        f = parse_field('xxx',name='yyy')
        self.assertIsNone(f.name)

    def test_parse_components(self):
        components = 'aaa^bbb'
        c = parse_components(components)
        self.assertEqual(c[0].to_er7(), 'aaa')
        self.assertEqual(c[1].to_er7(), 'bbb')
        components = '^comp^'
        c = parse_components(components)
        self.assertEqual(len(c), 3)
        self.assertEqual(c[0].to_er7(), '')
        self.assertEqual(c[1].to_er7(), 'comp')
        self.assertEqual(c[2].to_er7(), '')

    def test_parse_components_invalid_encoding_chars(self):
        components = 'aaa^bbb'
        self.assertRaises(InvalidEncodingChars, parse_components, components, encoding_chars=self._get_invalid_encoding_chars())
        self.assertRaises(InvalidEncodingChars, parse_components, components, encoding_chars=self._get_invalid_encoding_chars(),
                          validation_level = VALIDATION_LEVEL.STRICT)

    def test_parse_component(self):
        component = 'comp'
        c = parse_component(component)
        self.assertEqual(c.to_er7(), 'comp')
        component = '^comp^'
        c = parse_component(component)
        self.assertEqual(c.to_er7(), '\\S\\comp\\S\\')
        component = 'comp1&comp2'
        c = parse_component(component)
        self.assertEqual(c.to_er7(), 'comp1&comp2')
        self.assertEqual(c.children[0].to_er7(), 'comp1')
        self.assertEqual(c.children[1].to_er7(), 'comp2')

    def test_parse_component_invalid_encoding_chars(self):
        component = 'comp'
        self.assertRaises(InvalidEncodingChars, parse_component, component,
                          encoding_chars=self._get_invalid_encoding_chars())
        self.assertRaises(InvalidEncodingChars, parse_component, component,
                          encoding_chars=self._get_invalid_encoding_chars(),
                          validation_level = VALIDATION_LEVEL.STRICT)

    def test_parse_invalid_name_component_strict(self):
        self.assertRaises(InvalidName, parse_component, 'xxx', 'XXX', validation_level = VALIDATION_LEVEL.STRICT)

    def test_parse_subcomponents(self):
        subcomponent = 'subcomp'
        s = parse_subcomponents(subcomponent)
        self.assertEqual(s[0].to_er7(), 'subcomp')
        subcomponent = '&subcomp&'
        s = parse_subcomponents(subcomponent)
        self.assertEqual(s[0].to_er7(), '')
        self.assertEqual(s[1].to_er7(), 'subcomp')
        self.assertEqual(s[2].to_er7(), '')

    def test_parse_subcomponents_invalid_encoding_chars(self):
        subcomponents = 'sub1&subcomp&sub2'
        self.assertRaises(InvalidEncodingChars, parse_subcomponents, subcomponents,
                          encoding_chars=self._get_invalid_encoding_chars())
        self.assertRaises(InvalidEncodingChars, parse_subcomponents, subcomponents,
                          encoding_chars=self._get_invalid_encoding_chars(), validation_level = VALIDATION_LEVEL.STRICT)


if __name__ == '__main__':
    unittest.main()






# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2013, CRS4
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import socket

from actor import LB
from server import MLLProtocol, SB, EB, CR

def query(host, port, patient_id):
    """
    Establish a TCP connection to the given host and port and send a QBP_Q11 message
    :param host: address to connect to
    :param port: port to connect to
    :param patient_id: patient_id for the query message
    """
    # generate a QBP_Q11 message
    message = LB.send(patient_id)
    # establish the connection
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect((host, port))
        # send the message
        sock.sendall(SB + message + CR + EB + CR)
        # receive the answer
        received = sock.recv(1024*1024)
        message = MLLProtocol.get_message(received)
        # handle the response
        LB.receive(message)
    finally:
        sock.close()

if __name__ == '__main__':
    host, port = 'localhost', 6000
    patient_id = '100002'
    query(host, port, patient_id)
